/*
 * Copyright (C) 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.
 * Copyright (C) 2011 Google Inc. All rights reserved.
 * Copyright (C) 2013 ACCESS CO., LTD. All rights reserved.
 * Copyright (c) 2014 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(VIDEO)
#include "MediaControlRootElementWiiU.h"
#include "HTMLDivElement.h"

#include "Chrome.h" 
#include "ChromeClient.h" 
#include "Document.h" 
#include "HTMLMediaElement.h"
#include "HTMLNames.h"

#include "MediaControlElements.h"
#include "MouseEvent.h"
#include "Page.h"
#include "RenderTheme.h"
#include "Text.h"

#if PLATFORM(WIIU)
#include <PlatformUI.h>
#include <PlatformInput.h>
#include <PlatformConfig.h>
#include "KeyboardEvent.h"
#include "TimeRanges.h"
#endif
#if ENABLE(VIDEO_TRACK)
#include "TextTrackCue.h"
#endif

using namespace std;

extern LibPlatform::UI::UICallbacks_t* s_callbacks;

namespace WebCore {

static const double timeWithoutMouseMovementBeforeHidingControls = 15;


MediaControlEnclosureElement::MediaControlEnclosureElement(HTMLMediaElement* mediaElement)
    : MediaControlElement(mediaElement)
{
}

MediaControlElementType MediaControlEnclosureElement::displayType() const
{
    // Mapping onto same MediaControlElementType as panel element, since it has similar properties.
    return MediaControlsPanel;
}


MediaControlPanelEnclosureElement::MediaControlPanelEnclosureElement(HTMLMediaElement* mediaElement)
    : MediaControlEnclosureElement(mediaElement)
{
}

PassRefPtr<MediaControlPanelEnclosureElement> MediaControlPanelEnclosureElement::create(HTMLMediaElement* mediaElement)
{
    return adoptRef(new MediaControlPanelEnclosureElement(mediaElement));
}

const AtomicString& MediaControlPanelEnclosureElement::shadowPseudoId() const
{
    DEFINE_STATIC_LOCAL(AtomicString, id, ("-webkit-media-controls-enclosure"));
    return id;
}


MediaControlOverlayEnclosureElement::MediaControlOverlayEnclosureElement(HTMLMediaElement* mediaElement)
    : MediaControlEnclosureElement(mediaElement)
{
}

PassRefPtr<MediaControlOverlayEnclosureElement> MediaControlOverlayEnclosureElement::create(HTMLMediaElement* mediaElement)
{
    return adoptRef(new MediaControlOverlayEnclosureElement(mediaElement));
}

const AtomicString& MediaControlOverlayEnclosureElement::shadowPseudoId() const
{
    DEFINE_STATIC_LOCAL(AtomicString, id, ("-webkit-media-controls-overlay-enclosure"));
    return id;
}

MediaControlPauseNotificationOverlayEnclosureElement::MediaControlPauseNotificationOverlayEnclosureElement(HTMLMediaElement* mediaElement)
    : MediaControlEnclosureElement(mediaElement)
{
    hide();
}

PassRefPtr<MediaControlPauseNotificationOverlayEnclosureElement> MediaControlPauseNotificationOverlayEnclosureElement::create(HTMLMediaElement* mediaElement)
{
    return adoptRef(new MediaControlPauseNotificationOverlayEnclosureElement(mediaElement));
}

const AtomicString& MediaControlPauseNotificationOverlayEnclosureElement::shadowPseudoId() const
{
    DEFINE_STATIC_LOCAL(AtomicString, id, ("-webkit-media-controls-pause-overlay-enclosure"));
    return id;
}


MediaControlRootElement::MediaControlRootElement(HTMLMediaElement* mediaElement)
    : MediaControls(mediaElement)
    , m_mediaElement(mediaElement)
    , m_rewindButton(0)
    , m_playButton(0)
    , m_returnToRealTimeButton(0)
    , m_statusDisplay(0)
    , m_currentTimeDisplay(0)
    , m_timeline(0)
    , m_timeRemainingDisplay(0)
    , m_timelineContainer(0)
    , m_seekBackButton(0)
    , m_seekForwardButton(0)
    , m_toggleClosedCaptionsButton(0)
    , m_panelMuteButton(0)
    , m_volumeSlider(0)
    , m_volumeSliderMuteButton(0)
    , m_volumeSliderContainer(0)
    , m_fullScreenButton(0)
    , m_fullScreenMinVolumeButton(0)
    , m_fullScreenVolumeSlider(0)
    , m_fullScreenMaxVolumeButton(0)
    , m_panel(0)
    , m_enclosure(0)
#if ENABLE(VIDEO_TRACK)
    , m_textDisplayContainer(0)
#endif
    , m_hideControlsTimer(this, &MediaControlRootElement::hideControlsTimerFired)
    , m_fadeToBlackTimer(this, &MediaControlRootElement::fadeToBlackTimerFired)
    , m_pauseOverlayTimer(this, &MediaControlRootElement::pauseOverlayTimerFired)
    , m_isMouseOverControls(false)
    , m_isFullscreen(false)
    , m_overlayPlayButton(0)
    , m_overlayEnclosure(0)
    , m_playbackStartedOnce(false)
    , m_isTouchDown(false)
{
}

PassRefPtr<MediaControls> MediaControls::create(HTMLMediaElement* mediaElement)
{
    return MediaControlRootElement::create(mediaElement);
}

PassRefPtr<MediaControlRootElement> MediaControlRootElement::create(HTMLMediaElement* mediaElement)
{
	if (!mediaElement->document()->page())
		return 0;

    RefPtr<MediaControlRootElement> controls = adoptRef(new MediaControlRootElement(mediaElement));

    ExceptionCode ec;

    RefPtr<MediaControlOverlayEnclosureElement> overlayEnclosure = MediaControlOverlayEnclosureElement::create(mediaElement);
    RefPtr<MediaControlOverlayPlayButtonElement> overlayPlayButton = MediaControlOverlayPlayButtonElement::create(mediaElement);
    controls->m_overlayPlayButton = overlayPlayButton.get();
    overlayEnclosure->appendChild(overlayPlayButton.release(), ec, true);
    if (ec)
        return 0;

    RefPtr<MediaControlPauseNotificationOverlayEnclosureElement> pauseOverlayEnclosure = MediaControlPauseNotificationOverlayEnclosureElement::create(mediaElement);
    RefPtr<MediaControlPauseNotificationOverlayElement> pauseOverlay = MediaControlPauseNotificationOverlayElement::create(mediaElement);
    controls->m_overlayPauseNotification = pauseOverlay.get();
    pauseOverlayEnclosure->appendChild(pauseOverlay.release(), ec, true);
    controls->m_overlayPauseNotificationEnclosure = pauseOverlayEnclosure.get();
    if (ec)
        return 0;

    overlayEnclosure->appendChild(pauseOverlayEnclosure.release(), ec, true);
    if (ec)
        return 0;

    controls->m_overlayEnclosure = overlayEnclosure.get();
    controls->appendChild(overlayEnclosure.release(), ec, true);
    if (ec)
        return 0;

    if (controls->initializeControls(mediaElement))
        return controls.release();

    return 0;
}

bool MediaControlRootElement::initializeControls(HTMLMediaElement* mediaElement)
{
    // Create an enclosing element for the panel so we can visually offset the controls correctly.
    RefPtr<MediaControlPanelEnclosureElement> enclosure = MediaControlPanelEnclosureElement::create(mediaElement);

    RefPtr<MediaControlPanelElement> panel = MediaControlPanelElement::create(mediaElement);

    ExceptionCode ec;

    RefPtr<MediaControlRewindButtonElement> rewindButton = MediaControlRewindButtonElement::create(mediaElement);
    m_rewindButton = rewindButton.get();
    panel->appendChild(rewindButton.release(), ec, true);
    if (ec)
        return 0;

    RefPtr<MediaControlPlayButtonElement> playButton = MediaControlPlayButtonElement::create(mediaElement);
    m_playButton = playButton.get();
    panel->appendChild(playButton.release(), ec, true);
    if (ec)
        return 0;
        
    RefPtr<MediaControlReturnToRealtimeButtonElement> returnToRealtimeButton = MediaControlReturnToRealtimeButtonElement::create(mediaElement);
    m_returnToRealTimeButton = returnToRealtimeButton.get();
    panel->appendChild(returnToRealtimeButton.release(), ec, true);
    if (ec)
        return 0;

    if (document()->page()->theme()->usesMediaControlStatusDisplay()) {
        RefPtr<MediaControlStatusDisplayElement> statusDisplay = MediaControlStatusDisplayElement::create(mediaElement);
        m_statusDisplay = statusDisplay.get();
        panel->appendChild(statusDisplay.release(), ec, true);
        if (ec)
            return 0;
    }

    RefPtr<MediaControlTimelineContainerElement> timelineContainer = MediaControlTimelineContainerElement::create(mediaElement);

    RefPtr<MediaControlCurrentTimeDisplayElement> currentTimeDisplay = MediaControlCurrentTimeDisplayElement::create(mediaElement);
    m_currentTimeDisplay = currentTimeDisplay.get();
    timelineContainer->appendChild(currentTimeDisplay.release(), ec, true);
    if (ec)
        return 0;

    RefPtr<MediaControlTimelineElement> timeline = MediaControlTimelineElement::create(mediaElement, this);
    m_timeline = timeline.get();
    timelineContainer->appendChild(timeline.release(), ec, true);
    if (ec)
        return 0;

    RefPtr<MediaControlTimeRemainingDisplayElement> timeRemainingDisplay = MediaControlTimeRemainingDisplayElement::create(mediaElement);
    m_timeRemainingDisplay = timeRemainingDisplay.get();
    timelineContainer->appendChild(timeRemainingDisplay.release(), ec, true);
    if (ec)
        return 0;

    m_timelineContainer = timelineContainer.get();
    panel->appendChild(timelineContainer.release(), ec, true);
    if (ec)
        return 0;

    // FIXME: Only create when needed <http://webkit.org/b/57163>
    RefPtr<MediaControlSeekBackButtonElement> seekBackButton = MediaControlSeekBackButtonElement::create(mediaElement);
    m_seekBackButton = seekBackButton.get();
    panel->appendChild(seekBackButton.release(), ec, true);
    if (ec)
        return 0;

    // FIXME: Only create when needed <http://webkit.org/b/57163>
    RefPtr<MediaControlSeekForwardButtonElement> seekForwardButton = MediaControlSeekForwardButtonElement::create(mediaElement);
    m_seekForwardButton = seekForwardButton.get();
    panel->appendChild(seekForwardButton.release(), ec, true);
    if (ec)
        return 0;

    if (document()->page()->theme()->supportsClosedCaptioning()) {
        RefPtr<MediaControlToggleClosedCaptionsButtonElement> toggleClosedCaptionsButton = MediaControlToggleClosedCaptionsButtonElement::create(mediaElement);
        m_toggleClosedCaptionsButton = toggleClosedCaptionsButton.get();
        panel->appendChild(toggleClosedCaptionsButton.release(), ec, true);
        if (ec)
            return 0;
    }

        // FIXME: Only create when needed <http://webkit.org/b/57163>
    RefPtr<MediaControlFullscreenButtonElement> fullScreenButton = MediaControlFullscreenButtonElement::create(mediaElement, this);
        m_fullScreenButton = fullScreenButton.get();
    panel->appendChild(fullScreenButton.release(), ec, true);

    // The mute button and the slider element should be in the same div.
    RefPtr<HTMLDivElement> panelVolumeControlContainer = HTMLDivElement::create(document());

    if (document()->page()->theme()->usesMediaControlVolumeSlider()) {
        RefPtr<MediaControlVolumeSliderContainerElement> volumeSliderContainer = MediaControlVolumeSliderContainerElement::create(mediaElement);

        RefPtr<MediaControlVolumeSliderElement> slider = MediaControlVolumeSliderElement::create(m_mediaElement);
        m_volumeSlider = slider.get();
        volumeSliderContainer->appendChild(slider.release(), ec, true);
        if (ec)
            return 0;

        m_volumeSliderContainer = volumeSliderContainer.get();
        panelVolumeControlContainer->appendChild(volumeSliderContainer.release(), ec, true);
        if (ec)
            return 0;
    }

    RefPtr<MediaControlPanelMuteButtonElement> panelMuteButton = MediaControlPanelMuteButtonElement::create(mediaElement, this);
    m_panelMuteButton = panelMuteButton.get();
    panelVolumeControlContainer->appendChild(panelMuteButton.release(), ec, true);
    if (ec)
        return 0;

    panel->appendChild(panelVolumeControlContainer, ec, true);
    if (ec)
        return 0;

    // FIXME: Only create when needed <http://webkit.org/b/57163>
    RefPtr<MediaControlFullscreenVolumeMinButtonElement> fullScreenMinVolumeButton = MediaControlFullscreenVolumeMinButtonElement::create(mediaElement);
    m_fullScreenMinVolumeButton = fullScreenMinVolumeButton.get();
    panel->appendChild(fullScreenMinVolumeButton.release(), ec, true);
    if (ec)
        return 0;

    RefPtr<MediaControlFullscreenVolumeSliderElement> fullScreenVolumeSlider = MediaControlFullscreenVolumeSliderElement::create(mediaElement);
    m_fullScreenVolumeSlider = fullScreenVolumeSlider.get();
    panel->appendChild(fullScreenVolumeSlider.release(), ec, true);
    if (ec)
        return 0;

    RefPtr<MediaControlFullscreenVolumeMaxButtonElement> fullScreenMaxVolumeButton = MediaControlFullscreenVolumeMaxButtonElement::create(mediaElement);
    m_fullScreenMaxVolumeButton = fullScreenMaxVolumeButton.get();
    panel->appendChild(fullScreenMaxVolumeButton.release(), ec, true);
    if (ec)
        return 0;

    m_panel = panel.get();
    enclosure->appendChild(panel.release(), ec, true);
    if (ec)
        return false;

    m_enclosure = enclosure.get();
    appendChild(enclosure.release(), ec, true);
    if (ec)
        return false;

    return true;
}

//void MediaControlRootElement::setMediaController(MediaControllerInterface* controller)
//{
//    if (m_mediaController == controller)
//        return;
//
//    MediaControls::setMediaController(controller);
//
//    if(m_overlayPlayButton)
//    	 m_overlayPlayButton->setMediaController(controller);
//
//    if (m_rewindButton)
//        m_rewindButton->setMediaController(controller);
//    if (m_playButton)
//    	m_playButton->setMediaController(controller);
//    if (m_returnToRealTimeButton)
//        m_returnToRealTimeButton->setMediaController(controller);
//    if (m_statusDisplay)
//        m_statusDisplay->setMediaController(controller);
//    if (m_timeRemainingDisplay)
//        m_timeRemainingDisplay->setMediaController(controller);
//    if (m_timeline)
//    	m_timeline->setMediaController(controller);
//    if (m_timelineContainer)
//        m_timelineContainer->setMediaController(controller);
//    if (m_seekBackButton)
//        m_seekBackButton->setMediaController(controller);
//    if (m_seekForwardButton)
//        m_seekForwardButton->setMediaController(controller);
//    if (m_volumeSliderMuteButton)
//        m_volumeSliderMuteButton->setMediaController(controller);
//    if (m_volumeSliderContainer)
//        m_volumeSliderContainer->setMediaController(controller);
//    if (m_fullScreenMinVolumeButton)
//        m_fullScreenMinVolumeButton->setMediaController(controller);
//    if (m_fullScreenVolumeSlider)
//        m_fullScreenVolumeSlider->setMediaController(controller);
//    if (m_fullScreenMaxVolumeButton)
//        m_fullScreenMaxVolumeButton->setMediaController(controller);
//}

// Forces an update of the fullscreen controls so they don't show up in a "disabled" state 
//   when the video is paused while entering fullscreen.
void MediaControlRootElement::fixFullscreenControls()
{
    if(m_isFullscreen) {
        if (document() && document()->page()) {
            LibPlatform::UI::MediaShowHideControls(true, document()->page()->chrome()->client()->IsDisplayTV()); 
            LibPlatform::UI::MediaTogglePlayPause(!m_mediaElement->canPlay(), document()->page()->chrome()->client()->IsDisplayTV());
            RefPtr<TimeRanges> bufferedTimeRanges = m_mediaElement->buffered();
            ExceptionCode ignoredException;
            float end = bufferedTimeRanges->end(bufferedTimeRanges->length() - 1, ignoredException);
            LibPlatform::UI::MediaSetTimeLine(m_mediaElement->currentTime(), m_mediaElement->duration(), end, document()->page()->chrome()->client()->IsDisplayTV());
        }
    } 
}

void MediaControlRootElement::show()
{
    fixFullscreenControls();
    if(m_mediaElement->controls() && !m_isFullscreen) {
        m_panel->show();
    }
}

void MediaControlRootElement::hide()
{
    if(m_isFullscreen) {
		if (document() && document()->page())
			LibPlatform::UI::MediaShowHideControls(false, document()->page()->chrome()->client()->IsDisplayTV());
	} else {
		m_panel->hide();
	}
}

void MediaControlRootElement::makeOpaque()
{
    m_opaque = true;
}

void MediaControlRootElement::makeTransparent()
{
    m_opaque = false;
}

bool MediaControlRootElement::isVisible(bool isTv)
{
    if (m_isFullscreen) {
        return LibPlatform::UI::MediaGetControlsVisible(isTv);
    } else {
        return m_panel->isShowing();
    }
    return false;
}

void MediaControlRootElement::reset()
{
    Page* page = document()->page();
    if (!page)
        return;

    updateStatusDisplay();
	
	if(m_playButton) {
		if (m_playbackStartedOnce || !m_mediaElement->isVideo())
			m_playButton->show();
		else
			m_playButton->hide();
	}

	if(m_fullScreenButton) {
		if (m_mediaElement->supportsFullscreen() && LibPlatform::ApplicationConfig::GetVideoControlsEnabled())
			m_fullScreenButton->show();
		else
			m_fullScreenButton->hide();
	}

    float duration = m_mediaElement->duration();
    if (isfinite(duration) || page->theme()->hasOwnDisabledStateHandlingFor(MediaSliderPart)) {
        m_timeline->setDuration(duration);
        m_timelineContainer->show();
        m_timeline->setPosition(m_mediaElement->currentTime());
        updateTimeDisplay();
    } else
        m_timelineContainer->hide();

    if (m_mediaElement->hasAudio() || page->theme()->hasOwnDisabledStateHandlingFor(MediaMuteButtonPart))
        m_panelMuteButton->show();
    else
        m_panelMuteButton->hide();

    if (m_volumeSlider)
        m_volumeSlider->setVolume(m_mediaElement->volume());

    if (m_toggleClosedCaptionsButton) {
        if (m_mediaElement->hasClosedCaptions())
            m_toggleClosedCaptionsButton->show();
        else
            m_toggleClosedCaptionsButton->hide();
    }

    m_playButton->updateDisplayType();

#if ENABLE(FULLSCREEN_API)
    if (m_fullScreenVolumeSlider)
        m_fullScreenVolumeSlider->setVolume(m_mediaElement->volume());

    if (m_isFullscreen) {
        if (m_mediaElement->movieLoadType() == MediaPlayer::LiveStream) {
            m_seekBackButton->hide();
            m_seekForwardButton->hide();
            m_rewindButton->show();
            m_returnToRealTimeButton->show();
        } else {
            m_seekBackButton->show();
            m_seekForwardButton->show();
            m_rewindButton->hide();
            m_returnToRealTimeButton->hide();
        }
    } else
#endif
    if (m_mediaElement->movieLoadType() != MediaPlayer::LiveStream) {
        m_returnToRealTimeButton->hide();
        m_rewindButton->show();
    } else {
        m_returnToRealTimeButton->show();
        m_rewindButton->hide();
    }

    makeOpaque();

    if (!m_playbackStartedOnce && m_mediaElement->isVideo()) {
        hide();
    } else if (!m_mediaElement->isVideo()) {
        show();
    }
}

void MediaControlRootElement::playbackStarted()
{
    if (!m_playbackStartedOnce) {
		m_playButton->show();
        m_playbackStartedOnce = true;
        m_overlayPlayButton->hide();
        show();
        // make sure recalc style.
        toParentMediaElement(this)->setNeedsStyleRecalc();
        startHideControlsTimer(false);
    }

    // Inline controls are staying up after resuming playback.
    // So if they are up and there is not currently a timer, set one.
    if (m_panel->isShowing())
        startHideControlsTimer(false);

    if (m_isFullscreen && document() && document()->page()) {
        LibPlatform::UI::MediaTogglePlayPause(!m_mediaElement->canPlay(), document()->page()->chrome()->client()->IsDisplayTV());
    }
	
    m_playButton->updateDisplayType();
    m_timeline->setPosition(m_mediaElement->currentTime());
    updateTimeDisplay();
}

static bool _hardwareKeysInUse = false;
static void SetHWKeysUse (bool usingHWKeys)
{
    _hardwareKeysInUse = usingHWKeys;
}

bool MediaControlRootElement::isHardwareKeysInUse()
{
    return _hardwareKeysInUse;
}

void MediaControlRootElement::forceTimelineEvents()
{
    m_timeline->setPosition(m_mediaElement->currentTime());
}

void MediaControlRootElement::playbackProgressed()
{
    m_timeline->setPosition(m_mediaElement->currentTime());
    updateTimeDisplay();
    
    if (!m_isMouseOverControls && m_mediaElement->hasVideo())
        makeTransparent();
        
    // if hardware keys are in use, we don't want to hide the controls
    if (_hardwareKeysInUse)
        stopHideControlsTimer();
        
    if (!m_isTouchDown && !_hardwareKeysInUse && !m_hideControlsTimer.isActive() && (isVisible(false) || isVisible(true)))
        startHideControlsTimer(true);
    
    removePauseOverlay();
}

void MediaControlRootElement::playbackStopped()
{
	if (m_isFullscreen && document() && document()->page()) {
        LibPlatform::UI::MediaTogglePlayPause(!m_mediaElement->canPlay(), document()->page()->chrome()->client()->IsDisplayTV());
    }
    m_playButton->updateDisplayType();
    m_timeline->setPosition(m_mediaElement->currentTime());
    updateTimeDisplay();
    makeOpaque();
    
    stopHideControlsTimer();
}

void MediaControlRootElement::updateTimeDisplay()
{
    if (m_isFullscreen) {
    	RefPtr<TimeRanges> bufferedTimeRanges = m_mediaElement->buffered();
    	ExceptionCode ignoredException;
    	Page* page = document()->page();
    	if (!page)
    		return;
    	float end = bufferedTimeRanges->end(bufferedTimeRanges->length() - 1, ignoredException);
    	LibPlatform::UI::MediaSetTimeLine(m_mediaElement->currentTime(), m_mediaElement->duration(), end, page->chrome()->client()->IsDisplayTV());
    	return;
    }

    float now = m_mediaElement->currentTime();
    float duration = m_mediaElement->duration();

    Page* page = document()->page();
    if (!page)
        return;

    // Allow the theme to format the time.
    ExceptionCode ec;
    m_currentTimeDisplay->setInnerText(page->theme()->formatMediaControlsCurrentTime(now, duration), ec);
    m_currentTimeDisplay->setCurrentValue(now);
    // Set current time to duration and set duration to 0 so that "current - duration" becomes total time, because we want to show total time instead of remaining time.
    m_timeRemainingDisplay->setInnerText(page->theme()->formatMediaControlsRemainingTime(duration, 0), ec);
    m_timeRemainingDisplay->setCurrentValue(now - duration);
}

void MediaControlRootElement::reportedError()
{
}

void MediaControlRootElement::updateStatusDisplay()
{
    if (m_statusDisplay)
        m_statusDisplay->update();
}

void MediaControlRootElement::loadedMetadata()
{
    if (m_statusDisplay && m_mediaElement->movieLoadType() == MediaPlayer::LiveStream)
        m_statusDisplay->hide();

    reset();
}

void MediaControlRootElement::changedClosedCaptionsVisibility()
{
    if (m_toggleClosedCaptionsButton)
        m_toggleClosedCaptionsButton->updateDisplayType();
}

void MediaControlRootElement::changedMute()
{
    m_panelMuteButton->changedMute();
    if (m_volumeSliderMuteButton)
        m_volumeSliderMuteButton->changedMute();
}

void MediaControlRootElement::changedVolume()
{
    if (m_volumeSlider)
        m_volumeSlider->setVolume(m_mediaElement->volume());
}

void MediaControlRootElement::enteredFullscreen()
{
    ExceptionCode ignoredException;
    m_maxWidth = m_mediaElement->style()->removeProperty("max-width", ignoredException);
    m_maxHeight = m_mediaElement->style()->removeProperty("max-height", ignoredException);
    
    m_isFullscreen = true;
    s_callbacks->mediaSetHWUse = SetHWKeysUse;

    m_panel->setCanBeDragged(false);
	m_panel->hide();
    //m_currentTimeDisplay->hide(); 
    //m_timeRemainingDisplay->hide(); 
	
    if (Page* page = document()->page())
        page->chrome()->setCursorHiddenUntilMouseMoves(true);

    startFadeToBlackTimer();
    startHideControlsTimer(true);
}

void MediaControlRootElement::exitedFullscreen()
{
    ExceptionCode ignoredException;
    if (m_maxWidth.length() > 0)
        m_mediaElement->style()->setProperty("max-width", m_maxWidth, ignoredException);
    if (m_maxHeight.length() > 0)
        m_mediaElement->style()->setProperty("max-height", m_maxHeight, ignoredException);

    m_isFullscreen = false;
    s_callbacks->mediaSetHWUse = NULL;
    _hardwareKeysInUse = false;
    updateTimeDisplay(); // update the time on the inline controls

    m_panel->setCanBeDragged(false);

    // We will keep using the panel, but we want it to go back to the standard position.
    // This will matter right away because we use the panel even when not fullscreen.
    // And if we reenter fullscreen we also want the panel in the standard position.
    m_panel->resetPosition();
    
    m_panel->show();
    
    stopHideControlsTimer();
}

void MediaControlRootElement::showVolumeSlider()
{
    if (!m_mediaElement->hasAudio())
        return;

    if (m_volumeSliderContainer)
        m_volumeSliderContainer->show();
}

bool MediaControlRootElement::shouldHideControls()
{
    // hide if the hover is not coming from a remote
    return !m_panel->hovered() && !m_opaque;
}

bool MediaControlRootElement::containsRelatedTarget(Event* event)
{
    if (!event->isMouseEvent())
        return false;
    EventTarget* relatedTarget = static_cast<MouseEvent*>(event)->relatedTarget();
    if (!relatedTarget)
        return false;
    return contains(relatedTarget->toNode());
}

bool MediaControlRootElement::mouseTouchesControls(Event* event)
{
    bool isTv = document()->page()->chrome()->client()->IsDisplayTV();
    int screenWidth, screenHeight;
    if (isTv)
        LibPlatform::ApplicationConfig::GetTVDisplayDimensions(screenWidth, screenHeight);
    else
        LibPlatform::ApplicationConfig::GetGPDisplayDimensions(screenWidth, screenHeight);    
    MouseEvent *e = static_cast<MouseEvent*>(event);
    return LibPlatform::UI::MediaHitsControls(isTv, static_cast<float>(e->x()) / screenWidth, static_cast<float>(e->y()) / screenHeight);
}

void MediaControlRootElement::defaultEventHandler(Event* event)
{
    MediaControls::defaultEventHandler(event);
    //static bool noTimerRearm = false;

    // while the remote is on screen, reset the hide timer so that the 15 second timer begins
    //   when the remote cursor goes off screen
    if (!m_opaque && event->type() == eventNames().mousemoveEvent) {
        if (LibPlatform::Input::GetInstance()->CheckRemotePointerPosition(0)) {
            show();
            startHideControlsTimer(true);
        }
    }
    // don't hide the controls while the gamepad touchscreen is receiving input
    else if (event->type() == eventNames().touchmoveEvent) {
        m_isTouchDown = true;
        stopHideControlsTimer();
    }
    else if (!m_opaque && event->type() == eventNames().touchendEvent) {
        m_isTouchDown = false;
        startHideControlsTimer(true);
    }
    else if (event->type() == eventNames().clickEvent) {
    	Node *node = event->target()->toNode();
    	if((node->isHTMLElement() &&
    	   (toHTMLElement(node)->shadowPseudoId() == "-webkit-media-slider-thumb"
    	   || toHTMLElement(node)->shadowPseudoId() == "-webkit-media-controls-panel"
    	   || toHTMLElement(node)->shadowPseudoId() == "-webkit-media-slider-container"
    	   || toHTMLElement(node)->shadowPseudoId() == "-webkit-slider-runnable-track"))) {
        	stopHideControlsTimer();
    	} else {
            bool isTv = document()->page()->chrome()->client()->IsDisplayTV();
            bool hitsControls = mouseTouchesControls(event);
            // toggle visibility if it is not currently up, or if you touched outside the control bar to hide it
            if ((!isVisible(isTv) || !hitsControls) && m_playbackStartedOnce)
                toggleHiddenState();
    	}
    }
    else if (event->isKeyboardEvent()) {
    	if (event->type() == eventNames().keypressEvent) {
            switch (static_cast<KeyboardEvent*>(event)->charCode()) {
                case '\r':
                	m_overlayPlayButton->defaultEventHandler(event);
                	break;
            }
    	}
    }
}

void MediaControlRootElement::startHideControlsTimer(bool overwriteFireTime)
{
    if (!overwriteFireTime && m_hideControlsTimer.isActive())
        return;
    m_hideControlsTimer.startOneShot(timeWithoutMouseMovementBeforeHidingControls);
}

void MediaControlRootElement::hideControlsTimerFired(Timer<MediaControlRootElement>*)
{
    // only exit early if hover is coming from the wii remote here
    if ((!shouldHideControls() && LibPlatform::Input::GetInstance()->CheckRemotePointerPosition(0)) || m_opaque)
        return;

    if (Page* page = document()->page())
        page->chrome()->setCursorHiddenUntilMouseMoves(true);

    hide();
}

void MediaControlRootElement::stopHideControlsTimer()
{
    if (m_hideControlsTimer.isActive())
        m_hideControlsTimer.stop();
}

static OSTime s_fadeStartTime = 0;
void MediaControlRootElement::startFadeToBlackTimer()
{
    if (!m_fadeToBlackTimer.isActive()) {
        //m_fadeToBlackTimer.startRepeating(0.016);
        m_fadeToBlackTimer.startOneShot(0.016); // give the element a frame to complete entering fullscreen
    }
}
void MediaControlRootElement::stopFadeToBlackTimer()
{
    if (m_fadeToBlackTimer.isActive())
        m_fadeToBlackTimer.stop();
}

void MediaControlRootElement::fadeToBlackTimerFired(Timer<MediaControlRootElement>*)
{
    if (m_isFullscreen) {
        // ChromeClientWiiU handles fading to black when entering fullscreen because
        //  that is the earliest point we know a fullscreen request has been made.
        // This is used as a delayed point that once we know we have a fullscreen element,
        //  we can safely remove the black screen.
        bool isTv = document()->page()->chrome()->client()->IsDisplayTV();
        LibPlatform::UI::MediaSetFade(isTv, 0.0f);
        stopFadeToBlackTimer();
    } else {
        stopFadeToBlackTimer();
    }
}

void MediaControlRootElement::showPausedOverlay()
{
    m_overlayPauseNotificationEnclosure->show();
    if (!m_pauseOverlayTimer.isActive())
        m_pauseOverlayTimer.startOneShot(1.5);
}

void MediaControlRootElement::removePauseOverlay()
{
    if (m_pauseOverlayTimer.isActive()) {
        m_pauseOverlayTimer.stop();
        m_overlayPauseNotificationEnclosure->hide();
    }
}

void MediaControlRootElement::pauseOverlayTimerFired(Timer<MediaControlRootElement>*)
{
    m_overlayPauseNotificationEnclosure->hide();
}

#if ENABLE(VIDEO_TRACK)
void MediaControlRootElement::createTextTrackDisplay()
{
    if (m_textDisplayContainer)
        return;

    RefPtr<MediaControlTextTrackContainerElement> textDisplayContainer = MediaControlTextTrackContainerElement::create(document());
    m_textDisplayContainer = textDisplayContainer.get();

    // Insert it before the first controller element so it always displays behind the controls.
    insertBefore(textDisplayContainer.release(), m_enclosure, ASSERT_NO_EXCEPTION, true);
}

void MediaControlRootElement::showTextTrackDisplay()
{
    if (!m_textDisplayContainer)
        createTextTrackDisplay();
    m_textDisplayContainer->show();
}

void MediaControlRootElement::hideTextTrackDisplay()
{
    if (!m_textDisplayContainer)
        createTextTrackDisplay();
    m_textDisplayContainer->hide();
}

void MediaControlRootElement::updateTextTrackDisplay()
{
    if (!m_textDisplayContainer)
        createTextTrackDisplay();

    m_textDisplayContainer->updateDisplay();

}
#endif

const AtomicString& MediaControlRootElement::shadowPseudoId() const
{
    DEFINE_STATIC_LOCAL(AtomicString, id, ("-webkit-media-controls"));
    return id;
}

void MediaControlRootElement::bufferingProgressed()
{
    // We only need to update buffering progress when paused, during normal
    // playback playbackProgressed() will take care of it.
    if (m_mediaElement->paused())
        m_timeline->setPosition(m_mediaElement->currentTime());
}

void MediaControlRootElement::toggleHiddenState()
{
	if (document() && document()->page()) {
        if(!isVisible(document()->page()->chrome()->client()->IsDisplayTV()))
        {
            show();
            startHideControlsTimer(true);
        } 
        else if (shouldHideControls() && !LibPlatform::Input::GetInstance()->CheckRemotePointerPosition(0))
        {
            hide();
        }
	}
}


}

#endif
