/*
 * Copyright (C) 2011 Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "config.h"

#if ENABLE(WEB_AUDIO) && PLATFORM(WIIU)

//#if USE(WEBAUDIO_WIIU)

#include "FFTFrame.h"

#include "VectorMath.h"

#include <libffmpeg-fft/ffmpeg_FFT_WiiU.h>
#include <shim_memory.h>

#include <ppc.h>
#include <ppc_ps.h>
#include <wkaudio/wkaudAudio.h>

#include <wtf/MathExtras.h>
#include <map>

namespace WebCore {


class FFTFrame::Impl
    {
public:
    Impl(unsigned fftSize, unsigned fftSizeLog2);
    Impl(const Impl &rhs);
    ~Impl();

    void doFFT(float* data);
    void doInverseFFT(float* data);
	void doConvolve(const FFTFrame& frame, const float *pInput, float *pOutput, float *pLastOverlap);
	void multiply(const FFTFrame::Impl& frame);

private:
    static void *_fftAlloc(unsigned int numBytes);
    static void _fftFree(void *pAddress);
    static void *_fftRealloc(void *pAddress, unsigned int newByteSize);

    static void Initialize();

    static unsigned long s_bInitialized;

    void InitializeInstance();

    void        *m_pHandle;

    unsigned    m_FFTSize;
    unsigned    m_log2FFTSize;

    struct FFTinstance {
        FFTinstance() : m_refCnt(0) {}

        void *      m_pHandle;
        unsigned    m_refCnt;
    };
    typedef std::map<unsigned,FFTinstance>    FFTmap;

	static FFTmap *m_fftCache;

public:
    AudioFloatArray m_realData;
    AudioFloatArray m_imagData;
    };

unsigned long FFTFrame::Impl::s_bInitialized = false;
FFTFrame::Impl::FFTmap *FFTFrame::Impl::m_fftCache;

//static
void *FFTFrame::Impl::_fftAlloc(unsigned int numBytes)
{
    void *pA = malloc(size_t(numBytes));
    ASSERT(pA);
    return(pA);
}

//static
void FFTFrame::Impl::_fftFree(void *pAddress)
{
    free(pAddress);
}

//static
void *FFTFrame::Impl::_fftRealloc(void *pAddress, unsigned int newByteSize)
{
    return(shim_Realloc(pAddress, size_t(newByteSize)));
}

//static
void FFTFrame::Impl::Initialize()
{
    if (!s_bInitialized)
    {
        ffmpeg_fft_init_config_wiiu cfg;
        cfg.mpMemCallbacks.Allocate        = _fftAlloc;
        cfg.mpMemCallbacks.AllocateAligned = NULL;
        cfg.mpMemCallbacks.Deallocate      = _fftFree;
        cfg.mpMemCallbacks.Reallocate      = _fftRealloc;
        _ffmpeg_FFT_Init_WiiU(&cfg);

        ADD_TO_STATIC_LOCAL(s_bInitialized);
        s_bInitialized = true;
        ADD_TO_STATIC_LOCAL(FFTFrame::Impl::m_fftCache);
        FFTFrame::Impl::m_fftCache = new FFTFrame::Impl::FFTmap;
    }
}

void FFTFrame::Impl::InitializeInstance()
{
    // Do we already have an FFT instance for the given size?
    std::pair<FFTmap::iterator, bool> res = m_fftCache->insert(FFTmap::value_type(m_log2FFTSize, FFTinstance()));
    if (res.second)
    {
        // No! Make one...
        _ffmpeg_FFT_Create_WiiU(&m_pHandle, m_log2FFTSize);
        FFTinstance & inst = res.first->second;
        inst.m_pHandle = m_pHandle;
        ++inst.m_refCnt;
    }
    else
    {  
        // Yes. Just reuse it to save cycles etc.
        FFTinstance & inst = res.first->second;
        ++inst.m_refCnt;
        m_pHandle = inst.m_pHandle;
    }
}

FFTFrame::Impl::Impl(unsigned fftSize, unsigned fftSizeLog2)
    : m_FFTSize(fftSize)
    , m_log2FFTSize(fftSizeLog2)
    , m_realData(fftSize / 2)
    , m_imagData(fftSize / 2)
{
    Initialize();
    InitializeInstance();
}

FFTFrame::Impl::Impl(const FFTFrame::Impl &rhs)
    : m_FFTSize(rhs.m_FFTSize)
    , m_log2FFTSize(rhs.m_log2FFTSize)
    , m_realData(rhs.m_FFTSize / 2)
    , m_imagData(rhs.m_FFTSize / 2)
{
    Initialize();
    InitializeInstance();
}

FFTFrame::Impl::~Impl()
{
    // All users of this particular FFT size gone?
    FFTmap::iterator it = m_fftCache->find(m_log2FFTSize);
    FFTinstance & inst = it->second;
    if (--inst.m_refCnt == 0)
    {
        // Kick it!
        m_fftCache->erase(it);
        _ffmpeg_FFT_Destroy_WiiU(m_pHandle);
    }
}

// ----------------------------------------------------------------------------------------------
/**
*/
void FFTFrame::Impl::doFFT(float* data)
{
#if 1
	wksupAudioBeginSingleThread();

    // Compute Forward transform & keep it in internal buffers...
    float *p = _ffmpeg_FFT_CalcForward_To_WiiU(m_pHandle, data, NULL, 0);

    // FIXME: see above comment in multiply() about scaling.
    const float scale = 2.0f;

	// Deinterleave & scale from internal buffers to our data buffers...
	float *pReal = m_realData.data();
	float *pImag = m_imagData.data();
    for (int i=m_FFTSize>>2; i>0; --i)
	{
		f32x2 tmp0 = __PS_MULS0F(__PSQ_L(p, 0, 0), scale);
		f32x2 tmp1 = __PS_MULS0F(__PSQ_L(p + 2, 0, 0), scale);
		p += 4;
		__PSQ_ST(pReal, __PS_MERGE00(tmp0, tmp1), 0, 0);
		__PSQ_ST(pImag, __PS_MERGE11(tmp0, tmp1), 0, 0);
		pReal += 2;
		pImag += 2;
    }

	wksupAudioEndSingleThread();
#else
wksupAudioBeginSingleThread();
    // Copy since processing is in-place.
    float* p = m_complexData.data();
    memcpy(p, data, sizeof(float) * m_FFTSize);

    // Compute Forward transform.
    _ffmpeg_FFT_CalcForward_WiiU(m_pHandle, p);

    // De-interleave to separate real and complex arrays.
    int len = m_FFTSize / 2;

    // FIXME: see above comment in multiply() about scaling.
    const float scale = 2.0f;

    for (int i = 0; i < len; ++i) {
        int baseComplexIndex = 2 * i;
        // m_realData[0] is the DC component and m_imagData[0] is the nyquist component
        // since the interleaved complex data is packed.
        m_realData[i] = scale * p[baseComplexIndex];
        m_imagData[i] = scale * p[baseComplexIndex + 1];
    }
wksupAudioEndSingleThread();
#endif
}

// ----------------------------------------------------------------------------------------------
/**
*/
void FFTFrame::Impl::doInverseFFT(float* data)
{
#if 1
	wksupAudioBeginSingleThread();

    // Compute inverse transform.
    float *p = _ffmpeg_FFT_CalcInverseFromSeparate_To_WiiU(m_pHandle, m_realData.data(), m_imagData.data(), NULL);

    // Scale so that a forward then inverse FFT yields exactly the original data.
    const float scale = 1.0f / m_FFTSize;
    for (int i=m_FFTSize>>2; i>0; --i)
	{
		f32x2 tmp0 = __PS_MULS0F(__PSQ_L(p, 0, 0), scale);
		f32x2 tmp1 = __PS_MULS0F(__PSQ_L(p + 2, 0, 0), scale);
		p += 4;
		__PSQ_ST(data, tmp0, 0, 0);
		__PSQ_ST(data + 2, tmp1, 0, 0);
		data += 4;
    }

	wksupAudioEndSingleThread();
#else
wksupAudioBeginSingleThread();
    // Prepare interleaved data.
    float* interleavedData = getUpToDateComplexData();

	#if PLATFORM(WIIU)
    // Compute inverse transform.
    _ffmpeg_FFT_CalcInverse_WiiU(m_pHandle, interleavedData);
	#endif

    // Scale so that a forward then inverse FFT yields exactly the original data.
    const float scale = 1.0f / m_FFTSize;
    VectorMath::vsmul(interleavedData, 1, &scale, data, 1, m_FFTSize);
wksupAudioEndSingleThread();
#endif
}

// ----------------------------------------------------------------------------------------------
/**
*/
void FFTFrame::Impl::multiply(const FFTFrame::Impl& frame)
{
    FFTFrame::Impl& frame1 = *this;
    FFTFrame::Impl& frame2 = const_cast<FFTFrame::Impl&>(frame);

    float* realP1 = frame1.m_realData.data();
    float* imagP1 = frame1.m_imagData.data();
    const float* realP2 = frame2.m_realData.data();
    const float* imagP2 = frame2.m_imagData.data();

    // Scale accounts the peculiar scaling of vecLib on the Mac.
    // This ensures the right scaling all the way back to inverse FFT.
    // FIXME: if we change the scaling on the Mac then this scale
    // factor will need to change too.
    float scale = 0.5f;

    // Multiply the packed DC/nyquist component
    realP1[0] *= scale * realP2[0];
    imagP1[0] *= scale * imagP2[0];

    // Complex multiplication
    unsigned halfSize = m_FFTSize / 2;

	if (halfSize)
    {
		float realResult = realP1[1] * realP2[1] - imagP1[1] * imagP2[1];
		float imagResult = realP1[1] * imagP2[1] + imagP1[1] * realP2[1];
		realP1[1] = scale * realResult;
		imagP1[1] = scale * imagResult;
		
		for (unsigned i=2; i<halfSize; i+=2)
		{
			f32x2 rp1 = __PSQ_L(&realP1[i], 0, 0);
			f32x2 rp2 = __PSQ_L(&realP2[i], 0, 0);
			f32x2 ip1 = __PSQ_L(&imagP1[i], 0, 0);
			f32x2 ip2 = __PSQ_L(&imagP2[i], 0, 0);
			__PSQ_ST(&realP1[i], __PS_MULS0F(__PS_MSUB(rp1, rp2, __PS_MUL(ip1, ip2)), scale), 0, 0);
			__PSQ_ST(&imagP1[i], __PS_MULS0F(__PS_MADD(rp1, ip2, __PS_MUL(ip1, rp2)), scale), 0, 0);
		}
    }
}

// ----------------------------------------------------------------------------------------------
/**
*/
void FFTFrame::Impl::doConvolve(const FFTFrame& frame, const float *pInput, float *pOutput, float *pLastOverlap)
{
#if 0
	doFFT((float*)pInput);
	multiply(frame);
	doInverseFFT(pOutput);

	unsigned halfSize = m_FFTSize >> 1;

	// Overlap-add 1st half from previous time
	VectorMath::vadd(pOutput, 1, pLastOverlap, 1, pOutput, 1, halfSize);

	memcpy(pLastOverlap, pOutput + halfSize, sizeof(float) * halfSize);
#else
	wksupAudioBeginSingleThread();

	const unsigned halfSize = m_FFTSize >> 1;
	static const f32x2 cnst_m1p1 = {-1.0f, 1.0f};

	//
	// FFT
	//

    // Compute Forward transform & keep it in internal buffers...
	// (we flag the input buffer as being all zero in the second half - safes on memory access!)
    float *p = _ffmpeg_FFT_CalcForward_To_WiiU(m_pHandle, pInput, NULL, 1);

	//
	// complex multiply with outside kernel
	// (inside the FFT output buffer, as this is likely in LC!)
	//
    const float* realP2 = frame.m_Impl->m_realData.data();
    const float* imagP2 = frame.m_Impl->m_imagData.data();

    // Multiply the packed DC/nyquist component
    p[0] *= realP2[0];
    p[1] *= imagP2[0];

	// On to the "normal" stuff...
	p[2] = p[2] * realP2[1] - p[3] * imagP2[1];
	p[3] = p[2] * imagP2[1] + p[3] * realP2[1];

	float *pDst = &p[4];
	for (unsigned i=2; i<halfSize; i+=2)
	{
		f32x2 rp2 = __PSQ_L(&realP2[i], 0, 0);
		f32x2 ip2 = __PSQ_L(&imagP2[i], 0, 0);

		f32x2 c0 = __PSQ_L(&pDst[0], 0, 0);
		f32x2 c1 = __PSQ_L(&pDst[2], 0, 0);

		f32x2 cf0 = __PS_MERGE00(rp2, ip2);
		f32x2 cf1 = __PS_MERGE11(rp2, ip2);
		f32x2 cf0rev_m1p1 = __PS_MUL(__PS_MERGE00(ip2, rp2), cnst_m1p1);
		f32x2 cf1rev_m1p1 = __PS_MUL(__PS_MERGE11(ip2, rp2), cnst_m1p1);

		__PSQ_ST(&pDst[0], __PS_MADDS0(cf0, c0, __PS_MULS1(cf0rev_m1p1, c0)), 0, 0);
		__PSQ_ST(&pDst[2], __PS_MADDS0(cf1, c1, __PS_MULS1(cf1rev_m1p1, c1)), 0, 0);
		pDst += 4;
    }

	//
	// inverse FFT
	//

    // Compute inverse transform.
	// (note: we feed data from the buffer we got returned from the FFT code above! Luckily for us this is not the same as the input buffers!)
    p = _ffmpeg_FFT_CalcInverse_To_WiiU(m_pHandle, p, NULL);

    // 1st half: Scale so that a forward then inverse FFT yields exactly the original data & add with overlap buffer, writing into output buffer
	float *pOut = pOutput;
	const float *pLap = pLastOverlap;
    const float scale = 1.0f / m_FFTSize;
    for (int i=halfSize>>2; i>0; --i)
	{
		f32x2 tmp0 = __PS_MADDS0F(__PSQ_L(p, 0, 0), scale, __PSQ_L(pLap, 0, 0));
		f32x2 tmp1 = __PS_MADDS0F(__PSQ_L(p + 2, 0, 0), scale, __PSQ_L(pLap + 2, 0, 0));
		p += 4;
		pLap += 4;
		__PSQ_ST(pOut, tmp0, 0, 0);
		__PSQ_ST(pOut + 2, tmp1, 0, 0);
		pOut += 4;
    }
    // 2nd half: Scale so that a forward then inverse FFT yields exactly the original data & store in overlap buffer
	pOut = pLastOverlap;
    for (int i=halfSize>>2; i>0; --i)
	{
		f32x2 tmp0 = __PS_MULS0F(__PSQ_L(p, 0, 0), scale);
		f32x2 tmp1 = __PS_MULS0F(__PSQ_L(p + 2, 0, 0), scale);
		p += 4;
		__PSQ_ST(pOut, tmp0, 0, 0);
		__PSQ_ST(pOut + 2, tmp1, 0, 0);
		pOut += 4;
    }

	wksupAudioEndSingleThread();
#endif
}


// ----------------------------------------------------------------------------------------------
/**
	Normal constructor: allocates for a given fftSize.
*/
FFTFrame::FFTFrame(unsigned fftSize)
    : m_FFTSize(fftSize)
    , m_log2FFTSize(static_cast<unsigned>(log2(fftSize)))
{
    // We only allow power of two.
    ASSERT(1UL << m_log2FFTSize == m_FFTSize);
    m_Impl = new Impl(m_FFTSize, m_log2FFTSize);
}

// Creates a blank/empty frame (interpolate() must later be called).
FFTFrame::FFTFrame()
    : m_FFTSize(0)
    , m_log2FFTSize(0)
    , m_Impl(0)
{
}

// Copy constructor.
FFTFrame::FFTFrame(const FFTFrame& frame)
    : m_FFTSize(frame.m_FFTSize)
    , m_log2FFTSize(frame.m_log2FFTSize)
    , m_Impl(0)
{
    if (frame.m_Impl)
    {
        m_Impl = new Impl(*frame.m_Impl);
        // Copy/setup frame data.
        unsigned nbytes = sizeof(float) * (m_FFTSize / 2);
        memcpy(realData(), frame.realData(), nbytes);
        memcpy(imagData(), frame.imagData(), nbytes);
    }
}

void FFTFrame::initialize()
{
}

void FFTFrame::cleanup()
{
}

FFTFrame::~FFTFrame()
{
    delete m_Impl;
}

void FFTFrame::multiply(const FFTFrame& frame)
{
	m_Impl->multiply(*frame.m_Impl);
}

void FFTFrame::doFFT(float* data)
{
    m_Impl->doFFT(data);
}

void FFTFrame::doInverseFFT(float* data)
{
    m_Impl->doInverseFFT(data);
}

void FFTFrame::doConvolve(const FFTFrame& frame, const float *pInput, float *pOutput, float *pLastOverlap)
{
    m_Impl->doConvolve(frame, pInput, pOutput, pLastOverlap);
}

float* FFTFrame::realData() const
{
    return const_cast<float*>(m_Impl->m_realData.data());
}

float* FFTFrame::imagData() const
{
    return const_cast<float*>(m_Impl->m_imagData.data());
}

} // namespace WebCore

//#endif // USE(WEBAUDIO_WIIU)

#endif // ENABLE(WEB_AUDIO)
