/*
 * Copyright (C) 2006, 2007, 2008 Apple Inc.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2007 Pioneer Research Center USA, Inc.
 * Copyright (C) 2010 Igalia S.L.
 * Copyright (c) 2012 NINTENDO.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef FontPlatformDataFreeType_h
#define FontPlatformDataFreeType_h

#include "NotImplemented.h"
#define GLYPH_CACHE_SIZE 1024

#include "FontDescription.h"
#include "FontOrientation.h"
#include "GlyphBuffer.h"
#include <wtf/Forward.h>
#include <wtf/HashFunctions.h>

#include "ImageSource.h"

#include <vector>
#include <wtf/text/StringHash.h>

typedef struct glyph_cache_entry {
	short			bitmap_left;
	short			bitmap_top;
	short			bitmap_width;
	short			bitmap_rows;
	float			advance_x;
	float			s0, t0, s1, t1;
} glyph_cache_entry_t;

typedef struct font_cache_info {
	int				width;
	int				height;
	int				xpos;
	int				ypos;
	int				maxGlyphHeight;
} font_cache_info_t;

struct font_cache_pair_t {
	font_cache_pair_t() {}
	font_cache_pair_t(WebCore::Glyph g, int i) : mGlyph(g), mIndex(i) {}
	font_cache_pair_t(const font_cache_pair_t & src) : mGlyph(src.mGlyph), mIndex(src.mIndex) {}

	bool operator < (const font_cache_pair_t & rhs) const { return(mGlyph < rhs.mGlyph); }
	bool operator > (const font_cache_pair_t & rhs) const { return(mGlyph > rhs.mGlyph); }
	bool operator == (const font_cache_pair_t & rhs) const { return(mGlyph == rhs.mGlyph); }
	bool operator != (const font_cache_pair_t & rhs) const { return(mGlyph != rhs.mGlyph); }

	WebCore::Glyph	mGlyph;
	int				mIndex;
};

struct font_param {
	char 			*family_name;
	long 			style_flags;
	unsigned short  *characters;
	int 			num_chars;
    int 			font_weight;
};

struct font_config_handle {
	int not_used;
};

typedef std::vector<font_cache_pair_t>	font_cache_map_t;

struct font_face_t {
	static font_face_t* create();
	void destroy();
	void use();

    void*				face;
    bool				face_valid;

    font_cache_info_t	cache_info;
    glyph_cache_entry_t	glyph_cache[GLYPH_CACHE_SIZE];
	font_cache_map_t	glyph_map;
	WebCore::NativeImagePtr	m_FontSheet;

	unsigned int m_FontTextureID;
	unsigned int m_FontTextureWidth;
	unsigned int m_FontTextureHeight;

private:
	font_face_t();
	~font_face_t();
	unsigned int		ref_cnt;
};

inline font_face_t *font_face_t::create() {	return new font_face_t(); }
inline void font_face_t::destroy()		  {	if (--ref_cnt == 0) delete this; }
inline void font_face_t::use()			  {	++ref_cnt; }

namespace WebCore {

class FontPlatformData {
public:
    FontPlatformData(WTF::HashTableDeletedValueType)
        : m_size(0)
        , m_syntheticBold(false)
        , m_syntheticOblique(false)
        , m_scaledFont(0)
        , m_handle(NULL)
        {
    		memset(&m_pattern, 0, sizeof(m_pattern));
    		m_pattern.family_name = hashTableDeletedFontValue();
        }

    FontPlatformData()
        : m_size(0)
        , m_syntheticBold(false)
        , m_syntheticOblique(false)
        , m_scaledFont(0)
        , m_handle(NULL)
        {
            memset(&m_pattern, 0, sizeof(m_pattern));
            m_pattern.family_name = hashTableDeletedFontValue();
        }

    FontPlatformData(struct font_config_handle *handle, struct font_param *pattern, const FontDescription& fontDescription);
    FontPlatformData(font_face_t *fontFace, char * familyName, float size, bool bold, bool italic);
    FontPlatformData(float size, bool bold, bool italic);
    FontPlatformData(const FontPlatformData&);
    FontPlatformData(const FontPlatformData&, float size);
    ~FontPlatformData();

    bool isFixedPitch();
    float size() const { return m_size; }
    void setSize(float size) { m_size = size; }
    bool syntheticBold() const { return m_syntheticBold; }
    bool syntheticOblique() const { return m_syntheticOblique; }
    bool hasCompatibleCharmap();

    FontOrientation orientation() const { return Horizontal; } // FIXME: Implement.
    void setOrientation(FontOrientation) { } // FIXME: Implement.

    unsigned hash() const
    {
		if(m_pattern.family_name)
			return CaseFoldingHash::hash(m_pattern.family_name, strlen(m_pattern.family_name));
		else
			return 0;
    }

    bool operator==(const FontPlatformData&) const;
    FontPlatformData& operator=(const FontPlatformData&);
    bool isHashTableDeletedValue() const
    {
        return m_pattern.family_name ==  hashTableDeletedFontValue();
    }

#ifndef NDEBUG
    String description() const;
#endif

    float m_size;
    bool m_syntheticBold;
    bool m_syntheticOblique;
    bool m_fixedWidth;
    font_face_t *				m_scaledFont;
    struct font_config_handle *	m_handle;
    struct font_param			m_pattern;

    void createFontCache();
    int addGlyphToCache(Glyph glyph_idx) const;
	int getGlyphCacheIndex(Glyph glyph_idx) const;

private:
    static char *hashTableDeletedFontValue() { return reinterpret_cast<char *>(-1); }
};

}

#endif // FontPlatformDataFreeType_h
