/*
 * Copyright (C) 2006 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) 2007 Holger Hans Peter Freyther <zecke@selfish.org>
 * Copyright (C) 2008, 2009 Dirk Schulze <krit@webkit.org>
 * Copyright (C) 2010 Torch Mobile (Beijing) Co. Ltd. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ImageBuffer.h"

#include "Base64.h"
#include "BitmapImage.h"
#include "Color.h"
#include "GraphicsContext.h"
#include "GraphicsContextPlatformPrivate.h"
#include "ImageData.h"
#include "MIMETypeRegistry.h"
#include "NotImplemented.h"
#include "Pattern.h"
#include "PlatformString.h"
#include <wtf/Vector.h>
#include "StatisticsWiiU.h"

#include <png.h>

using namespace std;
using namespace WTF;

#if PLATFORM(WIIU)
#include <PlatformOS.h>
#include <PlatformMemory.h>
using namespace LibPlatform;
#endif

namespace WebCore {

// --------------------------------------------------------------------------------------------------
/**
*/
ImageBufferData::ImageBufferData(const IntSize& size) 
{
    m_platformContext = new GraphicsContextPlatformPrivate;
    m_platformContext->setSize(size.width(), size.height());
}

// --------------------------------------------------------------------------------------------------
/**
*/
ImageBuffer::ImageBuffer(const IntSize& size, ColorSpace, RenderingMode, bool& success, int aaMode, bool useMem1, int disableStencil, int format)
    : m_data(size)
    , m_size(size)
    , m_bDummyImageBuffer(false)
{ 
	GraphicsContext *ptr = new GraphicsContext(m_data.m_platformContext, aaMode, useMem1, disableStencil, format);
	if(ptr == NULL || ptr->isValid() == false)
	{
		if(ptr)	// this is annoying, but...
			delete ptr;
		success = false;
		return;
	}
    m_context = adoptPtr(ptr); 
    success = true;
	STATS_WIIU_INC_IMAGEBUFFER_COUNT();
}

ImageBuffer::ImageBuffer(const IntSize& size, GraphicsContext* context)
    : m_data(size)
    , m_size(size)
    , m_bDummyImageBuffer(true)
{
    m_context = adoptPtr(context);
    STATS_WIIU_INC_IMAGEBUFFER_COUNT();
}

// --------------------------------------------------------------------------------------------------
/**
*/
ImageBuffer::~ImageBuffer()
{
    //m_platformContext is freed by the graphics context platfromDestroy() 
    if (m_bDummyImageBuffer)
	{
        void *dummyP = m_context.leakPtr();
	}
	STATS_WIIU_DEC_IMAGEBUFFER_COUNT();
}

// --------------------------------------------------------------------------------------------------
/**
*/
size_t ImageBuffer::dataSize() const
{
	return m_context->platformContext()->getTextureDataSize();
}

// --------------------------------------------------------------------------------------------------
/**
*/
GraphicsContext* ImageBuffer::context() const
{
    return m_context.get();
}

// --------------------------------------------------------------------------------------------------
/**
	Make a copy of the render target to a BitmapImage
	
	@note	We do not support DontCopyBackingStore, which does not really copy the data, but hands out a reference of sort
*/
PassRefPtr<Image> ImageBuffer::copyImage(BackingStoreCopy copyBehavior) const
{
	// note: to support a non-copy reference, we'd need to change the native image type currently used as it's now only a HW texture with no room for ref-counts etc.
	ASSERT(copyBehavior == CopyBackingStore);
	
	NativeImagePtr texture = m_context->platformContext()->createTextureDataFromContext();
	PassRefPtr<WebCore::BitmapImage> bitmap_image = BitmapImage::create(texture, 0); 
	bitmap_image.get()->setSize(m_size.width(), m_size.height());
	return bitmap_image;
}

// --------------------------------------------------------------------------------------------------
/**
*/
void ImageBuffer::clip(GraphicsContext* context, const FloatRect& maskRect) const
{
#if PLATFORM(WIIU)
    OS::Report("ImageBuffer clip - NOT IMPLEMENTED");
#else
    printf("ImageBuffer clip - NOT IMPLEMENTED");
#endif
	notImplemented();
}

// --------------------------------------------------------------------------------------------------
/**
	Draw the contents of the context associated with this ImageBuffer to the context given
*/
void ImageBuffer::draw(GraphicsContext* context, ColorSpace styleColorSpace, const FloatRect& dst, const FloatRect& src,
                       CompositeOperator op , bool useLowQualityScale)
{
	// only supports this operation if the contexts are different!
	if (context == m_context)
		return;
	
	GraphicsContextPlatformPrivate *priv = m_context->platformContext();
	GraphicsContextPlatformPrivate *priv2 = context->platformContext();
	// Prep state of source context so we can use it as a texture
	if (!priv->prepareForUseAsTexture())
		return;

        
    if (priv2->isShadowed()) {

        ShadowBlur &shadow = priv2->getShadowBlurHandler();
        context->beginTransparencyLayer(context->getAlpha());
        if (shadow.mustUseShadowBlur(context)) {
            if (GraphicsContext *shadowContext = shadow.beginShadowLayer(context, dst)) {
                ColorSpace cs = ColorSpaceDeviceRGB;
                shadowContext->setAlpha(context->getAlpha());
                shadowContext->fillRect(dst, priv2->shadowColor(), cs);
                shadow.endShadowLayer(context);
            }
        } else {
            priv2->fillRect(WebCore::FloatRect(dst.x() + priv2->shadowWidth(), dst.y() + priv2->shadowHeight(), dst.width(), dst.height()), priv2->shadowColor());
           priv2->drawImage(priv->getTexture(),  WebCore::FloatRect(dst.x() + priv2->shadowWidth(), dst.y() + priv2->shadowHeight(), dst.width(), dst.height()), src, CompositeXOR);
  
        }
    }
    
	// draw it using standard code...
	context->platformContext()->drawImage(priv->getTexture(), dst, src, op);
    
    if (priv2->isShadowed())
        context->endTransparencyLayer();
}

// --------------------------------------------------------------------------------------------------
/**
	Draw the contents of the context associated with this ImageBuffer to the context given as a pattern
*/
void ImageBuffer::drawPattern(GraphicsContext* context, const FloatRect& srcRect, const AffineTransform& patternTransform,
                              const FloatPoint& phase, ColorSpace styleColorSpace, CompositeOperator op, const FloatRect& destRect)
{
	// Source in a different context then destination?
	if (context != m_context)
	{
		// Yes. We can do this without any copy-the-data action...
		if (!context->makeCurrent())
			return;
	
		// Make sure the texture unit sees the data from the source context
		m_context->invalidateBuffers();
	
		CompositeOperator oldOP = context->compositeOperation();
		context->setCompositeOperation(op);
		context->platformContext()->drawPattern(m_context->platformContext()->getTexture(), srcRect, phase, destRect, patternTransform);
		context->setCompositeOperation(oldOP);
	}
	else
	{
		// Same! We need to copy things and all...
		RefPtr<Image> image = copyImage(CopyBackingStore); 

		CompositeOperator oldOP = context->compositeOperation();
		context->setCompositeOperation(op);
		image->drawPattern(context, srcRect, patternTransform, phase, styleColorSpace, op, destRect);
		context->setCompositeOperation(oldOP);
	}
}

// --------------------------------------------------------------------------------------------------
/**
*/

void ImageBuffer::drawInstanced(GraphicsContext* context, unsigned int count, const float *pSrcRects, const float *pDestRects,
								const float *pMatrix32s, const float *pColors, ColorSpace styleColorSpace, CompositeOperator op)
{
	// only supports this operation if the contexts are different!
	ASSERT(context != m_context);
	if (context == m_context)
		return;
	
	GraphicsContextPlatformPrivate *priv = m_context->platformContext();
	
	// Prep state of source context so we can use it as a texture
	if (!priv->prepareForUseAsTexture())
		return;
	
	// draw it using standard code...
	context->platformContext()->drawImageInstanced(priv->getTexture(), count, pSrcRects, pDestRects, pMatrix32s, pColors, op);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void ImageBuffer::platformTransformColorSpace(const Vector<int>& lookUpTable)
{
	m_context->platformContext()->transformColorSpace(lookUpTable, m_size);
}

// --------------------------------------------------------------------------------------------------

PassRefPtr<ByteArray> ImageBuffer::getUnmultipliedImageData(const IntRect& rect) const
{
    return m_context->platformContext()->getUnmultipliedImageData(rect, m_size);
}

PassRefPtr<ByteArray> ImageBuffer::getPremultipliedImageData(const IntRect& rect) const
{
    return m_context->platformContext()->getPremultipliedImageData(rect, m_size);
}

// -------------------------------------------------------------------------------------------------

void ImageBuffer::putUnmultipliedImageData(ByteArray* source, const IntSize& sourceSize, const IntRect& sourceRect, const IntPoint& destPoint)
{
    m_context->platformContext()->putUnmultipliedImageData(source, sourceSize, sourceRect, destPoint, m_data, m_size);
}

void ImageBuffer::putPremultipliedImageData(ByteArray* source, const IntSize& sourceSize, const IntRect& sourceRect, const IntPoint& destPoint)
{
    m_context->platformContext()->putPremultipliedImageData(source, sourceSize, sourceRect, destPoint, m_data, m_size);
}

// --------------------------------------------------------------------------------------------------
/**
*/
static int dataurl_png_write(png_structp png, png_bytep data,
    png_size_t length)
{ 
    Vector<char> *in = (Vector<char> *)png_get_io_ptr(png);
    for (int i = 0; i < length; ++i) {
        in->append(data[i]);
    }

    return length; 
}

// --------------------------------------------------------------------------------------------------
/**
*/
String ImageBuffer::toDataURL(const String& mimeType, const double*) const
{ 
    String actualMimeType("image/png");
    if (MIMETypeRegistry::isSupportedImageMIMETypeForEncoding(mimeType))
        actualMimeType = mimeType;

    Vector<char> in;

    GraphicsContextPlatformPrivate *priv = m_context->platformContext();

#if PLATFORM(WIIU)
	int width, height, pitch;
	unsigned char* dataSrc = (unsigned char*)priv->getContextDataAsRGBA8(width, height, pitch);
	if (dataSrc)
	{
		png_structp png = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
		png_infop info = png_create_info_struct(png);
	
		//set up error handler
		if (setjmp(png_jmpbuf(png)))
		{
			png_destroy_write_struct(&png, &info);
			OS::Report("ImageBuffer::toDataURL > error could not write png to data url."); 
			return String();
		}
	
		png_set_write_fn(png, (void *)&in, (png_rw_ptr)dataurl_png_write, 0);
	
		png_set_IHDR(png, info, m_size.width(), m_size.height(), 8, PNG_COLOR_TYPE_RGBA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
	
		png_bytepp rows = (png_bytepp)Memory::AllocFromDefaultHeap(m_size.height()*sizeof(png_bytep));
		for (int i = 0; i < m_size.height(); ++i) {
			rows[i] = (png_bytep)(dataSrc + i*4*pitch);
		}
	
		png_write_info(png, info);
		png_write_image(png, rows); 
		png_write_end(png, 0); 
		png_destroy_write_struct(&png, &info);
	
		Memory::FreeToDefaultHeap(rows);
//note: we can free this right away, as we sync the GPU in the code above
		Memory::FreeToDefaultHeap(dataSrc);
	}

    Vector<char> out;
    base64Encode(in, out);

    return "data:" + actualMimeType + ";base64," + String(out.data(), out.size()); 
#else

	LOG_SIMPLE_RENDER_MSG("ImageBuffer::toDataURL()\n");

	// Setup a temporary color target in linear format and copy the current state of the color buffer into it with the GPU...
	GAMEPADERROR Result;
	unsigned int ImageWidth = priv->width();
	unsigned int ImageHeight = priv->height();
	unsigned int DataSize = ImageWidth * ImageHeight * 4;
	unsigned char * ImageBuffer = new unsigned char[DataSize];
	WiiURenderer * Renderer = priv->getRenderer();
	WiiURenderer::tRenderContext * CurrentContext = priv->getRenderContext();

	GraphicsContextPlatformPrivate::beginDraw();

	//GetOpenGLTextureData(ImageBuffer, 
	Result = GetOpenGLTextureData(ImageBuffer, CurrentContext->m_FBOInfo.m_ColorBufferID, ImageWidth, ImageHeight, true);
	if (Result)
	{
		ASSERT(false && "Could not read OpenGL texture data from image buffer");
		return "";
	}

	png_structp png = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	png_infop info = png_create_info_struct(png);

	//set up error handler
	if (setjmp(png_jmpbuf(png))) {
		png_destroy_write_struct(&png, &info);
		printf("ImageBuffer::toDataURL > error could not write png to data url.\n");
		return String();
	}

	png_set_write_fn(png, (void *)&in, (png_rw_ptr)dataurl_png_write, 0);

	png_set_IHDR(png, info, m_size.width(), m_size.height(), 8, PNG_COLOR_TYPE_RGBA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	png_bytepp rows = (png_bytepp)(new unsigned char [ImageHeight * sizeof(png_bytep)]);
	for (int i = 0; i < m_size.height(); ++i) {
		rows[i] = (png_bytep)(ImageBuffer + i * 4 * ImageWidth);
	}

	png_write_info(png, info);
	png_write_image(png, rows); 
	png_write_end(png, 0); 

	delete [] rows;
	delete [] ImageBuffer;

	Vector<char> out;
	base64Encode(in, out);

	return "data:" + actualMimeType + ";base64," + String(out.data(), out.size()); 
#endif

}

} // namespace WebCore
