/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2008 Kenneth Rohde Christiansen.  All rights reserved.
 * Copyright (C) 2009-2010 ProFUSION embedded systems
 * Copyright (C) 2009-2010 Samsung Electronics
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "Image.h"
#include "FloatRect.h"
#include "NotImplemented.h"

#include "BitmapImage.h"
#include "SharedBuffer.h"

#include "GraphicsContext.h"
#include "GraphicsContextPlatformPrivate.h"

#include "ImageResourceWiiU.h"
#include <wtf/text/StringConcatenate.h>

#include <wkgfx_Misc.h>
#include <wiiu/ImageDecoderWiiU.h>

using namespace WTF;

namespace WebCore {

void BitmapImage::initPlatformData()
{
	notImplemented();
}

void BitmapImage::invalidatePlatformData()
{
	notImplemented();
}

bool FrameData::clear(bool clearMetadata)
{
    if (clearMetadata)
        m_haveMetadata = false;

    if (m_frame)
	{
		ImageWiiU *imageP = (ImageWiiU*)m_frame;
		imageP->unref();
//		if(m_ignoreDestroyNativePtr == false)
//			WKgfx::DelayedNativeImageFree(m_frame);
        m_frame = NULL;
        return true;
    }

    return false;
}

BitmapImage::BitmapImage(void* surface, bool ignoreDestroyNativePtr, ImageObserver* observer)
    : Image(observer)
    , m_currentFrame(0)
    , m_frames(0)
    , m_frameTimer(0)
    , m_repetitionCount(cAnimationNone)
    , m_repetitionCountStatus(Unknown)
    , m_repetitionsComplete(0)
    , m_isSolidColor(false)
    , m_checkedForSolidColor(false)
    , m_animationFinished(true)
    , m_allDataReceived(true)
    , m_haveSize(true)
    , m_sizeAvailable(true)
    , m_hasUniformFrameSize(true)
    , m_decodedSize(0)
    , m_decodedPropertiesSize(0)
    , m_haveFrameCount(true)
    , m_frameCount(1)
{
    m_frames.grow(1);
    m_frames[0].m_frame = ImageWiiU::create(surface, ignoreDestroyNativePtr);
//	m_frames[0].m_ignoreDestroyNativePtr = ignoreDestroyNativePtr;	// for this constructor, the caller could be responsible for freeing this data

	//checkForSolidColor();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void BitmapImage::draw(GraphicsContext* context, const FloatRect& dst, const FloatRect& src, ColorSpace styleColorSpace, CompositeOperator op)
{
	startAnimation();
	
    NativeImagePtr texture = nativeImageForCurrentFrame();
    if (texture == NULL)
    	return;

   GraphicsContextPlatformPrivate *priv = context->platformContext();
    
   if (priv->isShadowed()) {
        
        ShadowBlur &shadow = priv->getShadowBlurHandler();
        context->beginTransparencyLayer(context->getAlpha()); 
        if (shadow.mustUseShadowBlur(context)) {
            if (GraphicsContext *shadowContext = shadow.beginShadowLayer(context, dst)) {
                ColorSpace cs = ColorSpaceDeviceRGB;
                shadowContext->setAlpha(context->getAlpha());
                shadowContext->fillRect(dst, priv->shadowColor(), cs);
                shadowContext->platformContext()->drawImage(texture, WebCore::FloatRect(dst.x() + priv->shadowWidth(), dst.y() + priv->shadowHeight(), dst.width(), dst.height()), src, CompositeXOR, false);
                shadow.endShadowLayer(context);
            }
        } else {
           priv->fillRect(WebCore::FloatRect(dst.x() + priv->shadowWidth(), dst.y() + priv->shadowHeight(), dst.width(), dst.height()), priv->shadowColor());
            priv->drawImage(texture, WebCore::FloatRect(dst.x() + priv->shadowWidth(), dst.y() + priv->shadowHeight(), dst.width(), dst.height()), src, CompositeXOR, false);            
        }
    }
    
	bool premultipliedAlpha = !m_IsGTX;
    priv->drawImage(texture, dst, src, op, premultipliedAlpha);
    if (priv->isShadowed())
        context->endTransparencyLayer(); 
}

// --------------------------------------------------------------------------------------------------
/**
*/
void BitmapImage::drawInstanced(GraphicsContext* context, unsigned int count, const float *pSrcRects, const float *pDestRects,
								const float *pMatrix32s, const float *pColors, ColorSpace styleColorSpace, CompositeOperator op)
{
	startAnimation();
	
    NativeImagePtr texture = nativeImageForCurrentFrame();
    if (texture == NULL)
    	return;

    bool premultipliedAlpha = !m_IsGTX;
    context->platformContext()->drawImageInstanced(texture, count, pSrcRects, pDestRects, pMatrix32s, pColors, op, premultipliedAlpha);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void Image::drawPattern(GraphicsContext* context, const FloatRect& tileRect, const AffineTransform& patternTransform,
                        const FloatPoint& phase, ColorSpace colorSpace, CompositeOperator op, const FloatRect& destRect)
{
    NativeImagePtr texture = nativeImageForCurrentFrame();
    if (texture == NULL)
		return;
	
    if (!context->makeCurrent())
		return;

	CompositeOperator oldOP = context->compositeOperation();
	context->setCompositeOperation(op);
	context->platformContext()->drawPattern(texture, tileRect, phase, destRect, patternTransform);
	context->setCompositeOperation(oldOP);
}


void BitmapImage::checkForSolidColor()
{
    m_isSolidColor = false;
    m_checkedForSolidColor = true;
    notImplemented();
}


static PassRefPtr<SharedBuffer> loadResourceSharedBufferFallback()
{
    return SharedBuffer::create(); // TODO: fallback image?
}

static PassRefPtr<SharedBuffer> loadResourceSharedBuffer(const char* name)
{
    RefPtr<SharedBuffer> buffer = SharedBuffer::createWithContentsOfFile(makeString(DATA_DIR "/webkit-1.0/images/", name, ".png"));    
    if (buffer)
        return buffer.release();
    return loadResourceSharedBufferFallback();
}

PassRefPtr<Image> Image::loadPlatformResource(const char* name)
{
    RefPtr<SharedBuffer> buffer;
    RefPtr<BitmapImage> img;

    if (strcmp(name, "textAreaResizeCorner") == 0)
       buffer =  SharedBuffer::create(ResizeCornerPng, sizeof (ResizeCornerPng));
    else if (strcmp(name, "textAreaResizeCorner@2x") == 0)
       buffer =  SharedBuffer::create(ResizeCornerPng2X, sizeof (ResizeCornerPng2X));
    else if (strcmp(name, "deleteButton") == 0)
       buffer =  SharedBuffer::create(DeleteButtonPng, sizeof (DeleteButtonPng));
    else if (strcmp(name, "deleteButton@2x") == 0)
       buffer =  SharedBuffer::create(DeleteButton2XPng, sizeof (DeleteButton2XPng));
    else if (strcmp(name, "missingImage") == 0)
       buffer =  SharedBuffer::create(MissingImagePng, sizeof (MissingImagePng));
    else if (strcmp(name, "missingImage@2x") == 0)
       buffer =  SharedBuffer::create(MissingImage2XPng, sizeof (MissingImage2XPng));
    else if (strcmp(name, "nullPlugin") == 0)
       buffer =  SharedBuffer::create(NullPluginPng, sizeof (NullPluginPng));
    else if (strcmp(name, "nullPlugin@2X") == 0)
       buffer =  SharedBuffer::create(NullPlugin2XPng, sizeof (NullPlugin2XPng));
    if (buffer) {
        img = BitmapImage::create();
        img->setData(buffer.release(), true);
        return img.release();
    }

    return Image::nullImage();
}


void *Image::getNativeTexture(void) {
    void *texture = nativeImageForCurrentFrame();
    if(texture == NULL) return 0;
    return texture;
}

}
