/*
 * Copyright (C) 2004, 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2007 Alp Toker <alp@atoker.com>
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2008 Collabora Ltd.
 * Copyright (C) 2008 Nuanti Ltd.
 * Copyright (C) 2009 Appcelerator Inc.
 * Copyright (C) 2009 Brent Fulgham <bfulgham@webkit.org>
 * Copyright (c) 2012 NINTENDO.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ResourceHandleManager.h"

#include "DataURL.h"
#include "HTTPParsers.h"
#include "MIMETypeRegistry.h"
#include "NotImplemented.h"
#include "ResourceError.h"
#include "ResourceHandle.h"
#include "ResourceHandleInternal.h"

#include <sys/stat.h>

#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
#include "DomainAccessControlWiiU.h"
#include "CredentialStorage.h"
#include "unistd.h"
#include "WebViewWiiU.h"
#include "wk_appinterface.h"
#include "PlatformNetwork.h"
#include "PlatformError.h"
#include "PlatformSSL.h"
#include "PlatformOS.h"
#include "ResourceLoader.h"
#include "TextEncoding.h"

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <shim.h>



#endif

#if !PLATFORM(WIIU)
#include "renderer/WiiURenderer.h"
#else
#define LOG_SIMPLE_RENDER_MSG(vMessage) {}
#define LOG_RENDER_MSG(vMessage, ...) {}
#endif

#include <errno.h>
#include <stdio.h>
#if USE(CF)
#include <wtf/RetainPtr.h>
#endif
#include <wtf/Threading.h>
#include <wtf/Vector.h>
#include <wtf/text/CString.h>
#include <wtf/StdLibExtras.h>

#if !OS(WINDOWS) && !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
#include <sys/param.h>
#define MAX_PATH MAXPATHLEN
#endif
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
#define INCREASE_DOWNLOAD_TIMER	(0)	// when this is 1, we care about total write size
#endif
#if PLATFORM(WIIU)
#define _NWF_SET_SOCKET_BUFFER_		// define to use large socket buffer for transfers
//#define _DISABLE_COOKIES				// define to disable all cookie jar access (for testing / for higher throughput with un-optimized file IO)
#endif

namespace WebCore {

const int selectTimeoutMS = 5;
const double pollTimeSeconds = 0.02;
const int maxRunningJobs = 5;

static const bool ignoreSSLErrors = getenv("WEBKIT_IGNORE_SSL_ERRORS");

class _xmlNode;


static void cleanupCacheResourceData(void *context)
{
	ResourceHandleManager::sharedInstance()->cleanupCacheResources();
	return;
}

static CString certificatePath()
{
	return LibPlatform::Network::GetCertificatePath();
}

static CString clientCertificatePath()
{
	return LibPlatform::Network::GetClientCertificatePath();
}

static CString crlPath()
{
	return LibPlatform::Network::GetCrlPath();
}

static Mutex* sharedResourceMutex(curl_lock_data data) {
    DEFINE_STATIC_LOCAL(Mutex, cookieMutex, ());
    DEFINE_STATIC_LOCAL(Mutex, dnsMutex, ());
    DEFINE_STATIC_LOCAL(Mutex, shareMutex, ());

    switch (data) {
        case CURL_LOCK_DATA_COOKIE:
            return &cookieMutex;
        case CURL_LOCK_DATA_DNS:
            return &dnsMutex;
        case CURL_LOCK_DATA_SHARE:
            return &shareMutex;
        default:
            ASSERT_NOT_REACHED();
            return NULL;
    }
}

// libcurl does not implement its own thread synchronization primitives.
// these two functions provide mutexes for cookies, and for the global DNS
// cache.
static void curl_lock_callback(CURL* handle, curl_lock_data data, curl_lock_access access, void* userPtr)
{
    if (Mutex* mutex = sharedResourceMutex(data))
        mutex->lock();
}

static void curl_unlock_callback(CURL* handle, curl_lock_data data, void* userPtr)
{
    if (Mutex* mutex = sharedResourceMutex(data))
        mutex->unlock();
}

ResourceHandleManager::ResourceHandleManager()
    : m_downloadTimer(this, &ResourceHandleManager::downloadTimerCallback)
    , m_cookieJarFileName(0)
    , m_certificatePath(certificatePath())
	, m_crlPath(crlPath())
    , m_clientCertificatePath(clientCertificatePath())
    , m_runningJobs(0)
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
	, m_cookieFlushTimer(this, &ResourceHandleManager::flushCookiesCallback)
    , m_readCacheTimer(this, &ResourceHandleManager::readCacheTimerCallback)
    , m_writeCacheTimer(this, &ResourceHandleManager::writeCacheTimerCallback) 
	, m_cookieLimit(-1)
#endif
{
    curl_global_init(CURL_GLOBAL_ALL);
    m_curlMultiHandle = curl_multi_init();
    m_curlShareHandle = curl_share_init();
    curl_share_setopt(m_curlShareHandle, CURLSHOPT_SHARE, CURL_LOCK_DATA_COOKIE);
    curl_share_setopt(m_curlShareHandle, CURLSHOPT_SHARE, CURL_LOCK_DATA_DNS);
    curl_share_setopt(m_curlShareHandle, CURLSHOPT_LOCKFUNC, curl_lock_callback);
    curl_share_setopt(m_curlShareHandle, CURLSHOPT_UNLOCKFUNC, curl_unlock_callback);
	m_increaseDownloadFrequency = false;
#if PLATFORM(WIIU)
	m_loadedCookies = false;
	parseCookieConfig();

    m_httpCache.parseConfigFile();
    m_httpsCache.parseConfigFile();
#endif
}

ResourceHandleManager::~ResourceHandleManager()
{
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
    forceTerminateHTTPCache();

	if (m_cookieFlushTimer.isActive()) {
		m_cookieFlushTimer.stop();
	}
	flushCookiesToDisk();
#endif
    curl_multi_cleanup(m_curlMultiHandle);
    curl_share_cleanup(m_curlShareHandle);
    if (m_cookieJarFileName)
        fastFree(m_cookieJarFileName);
    curl_global_cleanup();
}

void ResourceHandleManager::setCookieJarFileName(const char* cookieJarFileName)
{
#ifndef _DISABLE_COOKIES
    m_cookieJarFileName = fastStrDup(cookieJarFileName);
#endif
}

const char* ResourceHandleManager::getCookieJarFileName() const
{
#ifndef _DISABLE_COOKIES
	return m_cookieJarFileName;
#endif
}

CURLSH* ResourceHandleManager::getCurlShareHandle() const
{
    return m_curlShareHandle;
}

static ResourceHandleManager* sharedInstanceP = 0;

ResourceHandleManager* ResourceHandleManager::sharedInstance()
{
    if (!sharedInstanceP)
	{
        sharedInstanceP = new ResourceHandleManager();
		ADD_TO_STATIC_LOCAL(sharedInstanceP);
	}
    return sharedInstanceP;
}

void DeleteResourceHandleManager(void)
{
	if(sharedInstanceP)
		delete sharedInstanceP;
	sharedInstanceP = 0;
}

static void handleLocalReceiveResponse (CURL* handle, ResourceHandle* job, ResourceHandleInternal* d)
{
    // since the code in headerCallback will not have run for local files
    // the code to set the URL and fire didReceiveResponse is never run,
    // which means the ResourceLoader's response does not contain the URL.
    // Run the code here for local files to resolve the issue.
    // TODO: See if there is a better approach for handling this.
     const char* hdr;
     CURLcode err = curl_easy_getinfo(handle, CURLINFO_EFFECTIVE_URL, &hdr);
     ASSERT_UNUSED(err, CURLE_OK == err);

     d->m_response.setURL(KURL(ParsedURLString, hdr));
     if (d->client())
         d->client()->didReceiveResponse(job, d->m_response);
     d->m_response.setResponseFired(true);
}


// called with data after all headers have been processed via headerCallback
static size_t writeCallback(void* ptr, size_t size, size_t nmemb, void* data)
{ 
    ResourceHandle* job = static_cast<ResourceHandle*>(data);
    ResourceHandleInternal* d = job->getInternal();
    if (d->m_cancelled)
        return 0;

#if PLATFORM(WIIU) && !defined(BUILD_DEBUG) && !defined(HTTP_DEBUG)
	if(LibPlatform::OS::IsCodegenAvailable() == true)	// only HTTPS allowed...
	{
		// if this was not downloaded securely, make sure it is not executable code
		if (	job->firstRequest().url().protocol() != "https"
			&&  job->firstRequest().url().protocol() != "file")
		{
			String mimeType = d->m_response.httpHeaderField("Content-Type");
			if (!mimeType.isEmpty())
			{
				const String rgStrMediaMimeTypes[] = 
				{
					"image/",
					"audio/",
					"video/",
					"application/x-mpegURL",
					"vnd.apple.mpegURL"
				};

				const size_t numMediaMimeTypes = sizeof(rgStrMediaMimeTypes) / sizeof(*rgStrMediaMimeTypes);

				bool bIsMediaType = false;
				for (int mime = 0; mime < numMediaMimeTypes; mime++)
				{
					if (mimeType.startsWith(rgStrMediaMimeTypes[mime], false))
					{
						bIsMediaType = true;
						break;
					}
				}

				if (!bIsMediaType)
				{
					job->cancel();
					OS::Report("Must use HTTPS to download non-media file types.\n");
					OS::Report("Requested URL was: %s.\n", job->firstRequest().url().string().utf8().data());
					Error::GetInstance()->Report(Error::ERROR_CODE_CURLE_UNSUPPORTED_PROTOCOL, L"", 0);
					return 0;
				}
			}
			else
			{
				job->cancel();
				OS::Report("The Content-Type header must be set in order to download over an unsecured connection (non-HTTPS).\n");
				OS::Report("Requested URL was: %s.\n", job->firstRequest().url().string().utf8().data());
				Error::GetInstance()->Report(Error::ERROR_CODE_CURLE_UNSUPPORTED_PROTOCOL, L"", 0);
				return 0;
			}
		}
	}
#endif

#if LIBCURL_VERSION_NUM > 0x071200
    // We should never be called when deferred loading is activated.
    ASSERT(!d->m_defersLoading);
#endif

    size_t totalSize = size * nmemb;

#if INCREASE_DOWNLOAD_TIMER
	d->m_totalDownload += totalSize; 

	if(d->m_totalDownload > CURL_MAX_WRITE_SIZE) {
		ResourceHandleManager::sharedInstance()->decreaseDownloadFrequency();
		d->m_totalDownload = 0; 
	}
#endif

    // this shouldn't be necessary but apparently is. CURL writes the data
    // of html page even if it is a redirect that was handled internally
    // can be observed e.g. on gmail.com
    CURL* h = d->m_handle;
    long httpCode = 0;
    CURLcode err = curl_easy_getinfo(h, CURLINFO_RESPONSE_CODE, &httpCode);
    if (CURLE_OK == err && httpCode >= 300 && httpCode < 400)
        return totalSize;

    if (CURLE_OK == err && httpCode == 401 ) {
    	if (d->client()) {
    		if (d->m_authFailMSG && d->m_triedAuthentication) {
				//if authentication has failed, send the unauthorized text originally returned from the page
				if (d->m_cancelAuthentication)
				{
    				d->client()->didReceiveData(job, d->m_authFailMSG, d->m_authFailMSGSize, 0);
    				fastFree(d->m_authFailMSG);
					d->m_authFailMSG = NULL;
				}
    			return d->m_authFailMSGSize;
    		} else {
    			if (d->m_authFailMSG) {
    				fastFree(d->m_authFailMSG);
    			}
				//Copy the unauthorized text returned from the page
    			d->m_authFailMSG = (char*) fastMalloc(totalSize);
    			d->m_authFailMSGSize = totalSize;
    			memcpy(d->m_authFailMSG, static_cast<char*>(ptr), totalSize * sizeof(char));
    		}
    	}
    	return totalSize;
    }

    if (!d->m_response.responseFired()) {
        handleLocalReceiveResponse(h, job, d);
        if (d->m_cancelled)
            return 0;
    }

    if (d->client())
        d->client()->didReceiveData(job, static_cast<char*>(ptr), totalSize, 0);
    return totalSize;
}

/*
 * This is being called for each HTTP header in the response. This includes '\r\n'
 * for the last line of the header.
 *
 * We will add each HTTP Header to the ResourceResponse and on the termination
 * of the header (\r\n) we will parse Content-Type and Content-Disposition and
 * update the ResourceResponse and then send it away.
 *
 */
static size_t headerCallback(char* ptr, size_t size, size_t nmemb, void* data)
{
    ResourceHandle* job = static_cast<ResourceHandle*>(data);
    ResourceHandleInternal* d = job->getInternal();
    if (d->m_cancelled)
        return 0;

#if LIBCURL_VERSION_NUM > 0x071200
    // We should never be called when deferred loading is activated.
    ASSERT(!d->m_defersLoading);
#endif

    size_t totalSize = size * nmemb;
    ResourceHandleClient* client = d->client();

    String header(static_cast<const char*>(ptr), totalSize);

    /*
     * a) We can finish and send the ResourceResponse
     * b) We will add the current header to the HTTPHeaderMap of the ResourceResponse
     *
     * The HTTP standard requires to use \r\n but for compatibility it recommends to
     * accept also \n.
     */
    if (header == String("\r\n") || header == String("\n")) {
        CURL* h = d->m_handle;
        CURLcode err;

        double contentLength = 0;
        err = curl_easy_getinfo(h, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &contentLength);
        d->m_response.setExpectedContentLength(static_cast<long long int>(contentLength));

        const char* hdr;
        err = curl_easy_getinfo(h, CURLINFO_EFFECTIVE_URL, &hdr);
        d->m_response.setURL(KURL(ParsedURLString, hdr));

        long httpCode = 0;
        err = curl_easy_getinfo(h, CURLINFO_RESPONSE_CODE, &httpCode);
        d->m_response.setHTTPStatusCode(httpCode);

        d->m_response.setMimeType(extractMIMETypeFromMediaType(d->m_response.httpHeaderField("Content-Type")));
        d->m_response.setTextEncodingName(extractCharsetFromMediaType(d->m_response.httpHeaderField("Content-Type")));
        d->m_response.setSuggestedFilename(filenameFromHTTPContentDisposition(d->m_response.httpHeaderField("Content-Disposition")));

        // HTTP redirection
        if (httpCode >= 300 && httpCode < 400) {
            String location = d->m_response.httpHeaderField("location");
            if (!location.isEmpty()) {
                KURL newURL = KURL(job->firstRequest().url(), location);

                ResourceRequest redirectedRequest = job->firstRequest();
                redirectedRequest.setURL(newURL);
                if (client)
                    client->willSendRequest(job, redirectedRequest, d->m_response);

                d->m_firstRequest.setURL(newURL);

                return totalSize;
            }
        }

        if (client) {
			// Fixme: As far as I'm concerned this is a hack.  -Nsa-LH 6/20/2012
			//        When making https requests through a proxy, a CONNECT request is made
			//         to the proxy before making a SSL connection to the remote server.
			//        libcurl will intentionally call this callback with the proxy's
			//         response to the CONNECT, and this occurs in such a way that
			//         it appears before the remote server's response.
			//        Calling client->didReceiveResponse() at this point will cause the
			//         request to prematurely error as there was no actual header information
			//         at this point, so we're gonna avoid this call in the event that it
			//         was a proxy CONNECT response.
			if (httpCode == 0) {
				err = curl_easy_getinfo(h, CURLINFO_HTTP_CONNECTCODE, &httpCode);
				// Check the proxy connect code is HTTP OK and protocol is https
				// Don't specifically check if a proxy was set because libcurl respects
				//  global proxy settings from environment variables as well
				if (httpCode == 200 && protocolIs(String(hdr), "https"))
					return totalSize;
			}
#if PLATFORM(WIIU)
			else if (httpCode >= 400 && httpCode <= 599)
				Error::GetInstance()->Report((Error::ErrorCode)(Error::GetHTTPStatusCodesOffset() + httpCode), L"", 0);
#endif

            client->didReceiveResponse(job, d->m_response);
		}
        d->m_response.setResponseFired(true);

    } else {
        int splitPos = header.find(":");
        if (splitPos != -1)
		{
            d->m_response.setHTTPHeaderField(header.left(splitPos), header.substring(splitPos+1).stripWhiteSpace());
		}
    }

    return totalSize;
}

/* This is called to obtain HTTP POST or PUT data.
   Iterate through FormData elements and upload files.
   Carefully respect the given buffer size and fill the rest of the data at the next calls.
*/
size_t readCallback(void* ptr, size_t size, size_t nmemb, void* data)
{
    ResourceHandle* job = static_cast<ResourceHandle*>(data);
    ResourceHandleInternal* d = job->getInternal();

    if (d->m_cancelled)
        return 0;

#if LIBCURL_VERSION_NUM > 0x071200
    // We should never be called when deferred loading is activated.
    ASSERT(!d->m_defersLoading);
#endif

    if (!size || !nmemb)
        return 0;

    if (!d->m_formDataStream.hasMoreElements())
        return 0;

    size_t sent = d->m_formDataStream.read(ptr, size, nmemb);

    // Something went wrong so cancel the job.
    if (!sent)
        job->cancel();

    return sent;
}

void ResourceHandleManager::downloadTimerCallback(Timer<ResourceHandleManager>* timer)
{
    LOG_SIMPLE_RENDER_MSG("!!!!!!ResourceHandleManager::downloadTimerCallback()\n");
    startScheduledJobs();

    fd_set fdread;
    fd_set fdwrite;
    fd_set fdexcep;
    int maxfd = 0;
	int runningHandles = 0;
    int rc = 0;
    struct timeval timeout;

#if INCREASE_DOWNLOAD_TIMER

    timeout.tv_sec = 0;
    timeout.tv_usec = selectTimeoutMS * 1000;       // select waits microseconds

    LOG_SIMPLE_RENDER_MSG("    Do the things\n");
    // Retry 'select' if it was interrupted by a process signal.
    int rc = 0;
    do {
        FD_ZERO(&fdread);
        FD_ZERO(&fdwrite);
        FD_ZERO(&fdexcep);
        curl_multi_fdset(m_curlMultiHandle, &fdread, &fdwrite, &fdexcep, &maxfd);
        // When the 3 file descriptors are empty, winsock will return -1
        // and bail out, stopping the file download. So make sure we
        // have valid file descriptors before calling select.
        if (maxfd >= 0)
            rc = ::select(maxfd + 1, &fdread, &fdwrite, &fdexcep, &timeout);
    } while (rc == -1 && errno == EINTR);
#else
 
    /* we start some action by calling perform right away */
    curl_multi_perform(m_curlMultiHandle, &runningHandles);

    timeout.tv_sec = 0;
    timeout.tv_usec = 0; //selectTimeoutMS * 1000;       // select waits microseconds

    LOG_SIMPLE_RENDER_MSG("    Do the things\n");

    FD_ZERO(&fdread);
    FD_ZERO(&fdwrite);
    FD_ZERO(&fdexcep);

    /* set a suitable timeout to play around with */
    timeout.tv_sec = 0;
    timeout.tv_usec = 0; //selectTimeoutMS * 1000;

    curl_multi_fdset(m_curlMultiHandle, &fdread, &fdwrite, &fdexcep, &maxfd);
        // When the 3 file descriptors are empty, winsock will return -1
        // and bail out, stopping the file download. So make sure we
        // have valid file descriptors before calling select.
    if (maxfd >= 0)
        rc = ::select(maxfd + 1, &fdread, &fdwrite, &fdexcep, &timeout);
#endif
    LOG_SIMPLE_RENDER_MSG("    The things are did.\n");

    if (-1 == rc) {
#ifndef NDEBUG
        perror("bad: select() returned -1: ");
#endif
        return;
    }

    LOG_SIMPLE_RENDER_MSG("    Perform the curls.\n");
    int authChallengeCount = 0;
    bool firstTimeThrough = true;
    while (curl_multi_perform(m_curlMultiHandle, &runningHandles) == CURLM_CALL_MULTI_PERFORM) {
        if(m_loadedCookies == true && firstTimeThrough == true) {
        	curl_easy_setopt(m_curlMultiHandle, CURLOPT_COOKIEFILE, NULL);
			curl_easy_setopt(m_curlMultiHandle, CURLOPT_COOKIEJAR, NULL);
            firstTimeThrough = false;
        }
    }
    LOG_SIMPLE_RENDER_MSG("    Curls performed.\n");

    // check the curl messages indicating completed transfers
    // and free their resources


    LOG_SIMPLE_RENDER_MSG("    While the trues.\n");
    while (true) {
        int messagesInQueue;
        CURLMsg* msg = curl_multi_info_read(m_curlMultiHandle, &messagesInQueue);
        if (!msg)
            break;

        // find the node which has same d->m_handle as completed transfer
        CURL* handle = msg->easy_handle;
        ASSERT(handle);
        ResourceHandle* job = 0;
        CURLcode err = curl_easy_getinfo(handle, CURLINFO_PRIVATE, &job);
        ASSERT_UNUSED(err, CURLE_OK == err);
        ASSERT(job);
        if (!job)
            continue;
        ResourceHandleInternal* d = job->getInternal();
        ASSERT(d->m_handle == handle);

        if (d->m_cancelled) {
            removeFromCurl(job);
            continue;
        }

        if (CURLMSG_DONE != msg->msg)
            continue;

        if (CURLE_OK == msg->data.result) {
        	if (d->m_response.httpStatusCode() == 401) { // 401 is an authentication challenge
        		ASSERT(isMainThread());

        		if (d->m_triedAuthentication) {
        	           if (!d->m_response.responseFired()) {
        	                handleLocalReceiveResponse(d->m_handle, job, d);
        	                if (d->m_cancelled) {
        	                    removeFromCurl(job);
        	                    continue;
        	                }
        	            }

						if(!d->m_cancelAuthentication)
						{
							ProtectionSpaceAuthenticationScheme authScheme = ProtectionSpaceAuthenticationSchemeDefault;
        					const KURL& kurl = job->firstRequest().url(); // grab the URL
							String authString = d->m_response.httpHeaderField("WWW-Authenticate");
        					String realm;

        					if (authString.startsWith("Basic ", false)) {
        						authScheme = ProtectionSpaceAuthenticationSchemeHTTPBasic;
        						String realmStr = authString.substring(6);
        						if (realmStr.startsWith("realm=\"", false)) {
        							realm = realmStr.substring(7, realmStr.length()-8);
        						}
        					}
        					else if (authString.startsWith("Digest ", false)){
        						authScheme = ProtectionSpaceAuthenticationSchemeHTTPDigest;
        					}
        					else if (authString.startsWith("Negotiate", false)){
        						authScheme = ProtectionSpaceAuthenticationSchemeNegotiate;
        					}
        					const ProtectionSpace protectionSpace(kurl.host(), kurl.port(), ProtectionSpaceServerHTTP, realm, authScheme);
							//remove stored credential info
							CredentialStorage::remove(protectionSpace);
							d->m_pass = "";
							d->m_triedAuthentication = false;
						}
						else
						{
							if (d->client())
							{
								if(d->m_authFailMSG)
								{
									d->client()->didReceiveData(job, d->m_authFailMSG, d->m_authFailMSGSize, 0);
    								fastFree(d->m_authFailMSG);
									d->m_authFailMSG = NULL;
								}
        						d->client()->didFinishLoading(job, 0);
							}
        					continue;
						}
        		}

        		ProtectionSpaceAuthenticationScheme authScheme = ProtectionSpaceAuthenticationSchemeDefault;
        		const KURL& kurl = job->firstRequest().url(); // grab the URL

        		String authString = d->m_response.httpHeaderField("WWW-Authenticate");
        		String realm;

        		if (authString.isEmpty()) {
        			cancel(job);
        			d->m_initialCredential = Credential();
					if (d->client())
					{
						//Fake readystate 2-4
						d->client()->didReceiveData(job, " ", 1, 0);
        				d->client()->didFinishLoading(job, 0);
					}
        			return;
        		}

        		if (authString.startsWith("Basic ", false)) {
        			authScheme = ProtectionSpaceAuthenticationSchemeHTTPBasic;
        			String realmStr = authString.substring(6);
        			if (realmStr.startsWith("realm=\"", false)) {
        				realm = realmStr.substring(7, realmStr.length()-8);
        			}
        		}
        		else if (authString.startsWith("Digest ", false)){
        			authScheme = ProtectionSpaceAuthenticationSchemeHTTPDigest;
        		}
        		else if (authString.startsWith("Negotiate", false)){
        			authScheme = ProtectionSpaceAuthenticationSchemeNegotiate;
        		}
        		const ProtectionSpace protectionSpace(kurl.host(), kurl.port(), ProtectionSpaceServerHTTP, realm, authScheme);

        		Credential credential;

        		if (!d->m_user.isEmpty() || !d->m_pass.isEmpty()) {
        			credential = Credential(d->m_user, d->m_pass, WebCore::CredentialPersistenceForSession);
        			CredentialStorage::set(credential, protectionSpace, kurl);
        		}
        		else {
        			credential = CredentialStorage::get(protectionSpace);
        		}

        		d->m_currentWebChallenge = AuthenticationChallenge(protectionSpace, credential, authChallengeCount++, d->m_response, ResourceError());
				d->m_currentWebChallenge.m_sourceHandle = job;

        		if (d->client() && !credential.hasPassword()) {
        			d->client()->didReceiveAuthenticationChallenge(job, d->m_currentWebChallenge);
        			d->m_initialCredential = CredentialStorage::get(protectionSpace);
        		}
        		else {
        			d->m_initialCredential = credential;
        		}

        		if (d->m_initialCredential.hasPassword()){
        			String userpassword;
        			d->m_user = d->m_initialCredential.user();
        			d->m_pass = d->m_initialCredential.password();
        		}

        		if (d->m_cancelled) {
        			removeFromCurl(job);
        			continue;
        		}

        		if ((d->m_initialCredential.hasPassword() && ((d->m_currentWebChallenge.proposedCredential() != d->m_initialCredential) || (authChallengeCount <= 1)))
					|| d->m_cancelAuthentication)
        		{
        			job->ref();
        			removeFromCurl(job);
        			free(d->m_url);
        			d->m_url = NULL;
        			initializeHandle(job);
        			d->m_triedAuthentication = true;
        			switch (protectionSpace.authenticationScheme())
        			{
        			case ProtectionSpaceAuthenticationSchemeDefault:
        				curl_easy_setopt(d->m_handle, CURLOPT_HTTPAUTH, CURLAUTH_ANY);
        				break;
        			case ProtectionSpaceAuthenticationSchemeHTTPBasic:
        				curl_easy_setopt(d->m_handle, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        				break;
        			case ProtectionSpaceAuthenticationSchemeHTTPDigest:
        				curl_easy_setopt(d->m_handle, CURLOPT_HTTPAUTH, CURLAUTH_DIGEST | CURLAUTH_DIGEST_IE);
        				break;
        			case ProtectionSpaceAuthenticationSchemeNTLM:
        				curl_easy_setopt(d->m_handle, CURLOPT_HTTPAUTH, CURLAUTH_NTLM);
        				break;
        			case ProtectionSpaceAuthenticationSchemeNegotiate:
        				curl_easy_setopt(d->m_handle, CURLOPT_HTTPAUTH, CURLAUTH_GSSNEGOTIATE);
        				break;
        			}
        			CURLMcode ret = curl_multi_add_handle(m_curlMultiHandle, job->getInternal()->m_handle);
        			ASSERT(ret == CURLM_OK || ret == CURLM_CALL_MULTI_PERFORM);
        		    if (ret && ret != CURLM_CALL_MULTI_PERFORM) {
        		        job->cancel();
        		        return;
        		    }
        		    ++runningHandles;
        		    continue;
        		}
        	}

            if (!d->m_response.responseFired()) {
                handleLocalReceiveResponse(d->m_handle, job, d);
                if (d->m_cancelled) {
                    removeFromCurl(job);
                    continue;
                }
            }

            if (d->client()) {
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
                KURL url = job->firstRequest().url();

                HTTPCache *cache = 0;
                if (url.protocol() == "https") {
                    cache = &m_httpsCache; 
                } else {
                    cache = &m_httpCache; 
                }

                if (!cache->disabled()) {
                    switch (d->m_response.httpStatusCode()) {
                    case 200:
                    case 201:
                    case 300:
                    case 301:
                    case 410:
                        addHTTPCache(job, url, d->client()->resourceData(), d->m_response);
                        break;
                    case 304:
                        {
                            if (job->firstRequest().cachePolicy() == ReloadIgnoringCacheData)
                                break;  // read from the memory cached resource

                            // read from the disk cached resource
                            HTTPCachedResource *resource = cache->resourceForURL(url);
                            if (resource)
                                scheduleLoadResourceFromHTTPCache(job);
                            else
                                d->client()->didFail(job, ResourceError(String(), CURLE_READ_ERROR, String(d->m_url), String("cache read error")));
                            removeFromCurl(job);
                            continue;
                        }
                    default:
                        break;
                    }
                }
 #endif
                d->client()->didFinishLoading(job, 0);
			}
        } else {
            char* url = 0;
            curl_easy_getinfo(d->m_handle, CURLINFO_EFFECTIVE_URL, &url);
#ifndef NDEBUG
#if PLATFORM(WIIU)
			switch (msg->data.result)
			{
			case CURLE_COULDNT_RESOLVE_PROXY:
				Error::GetInstance()->Report(Error::ERROR_CODE_CURLE_COULDNT_RESOLVE_PROXY, L"", 0);
				break;

			case CURLE_SSL_CACERT:
				if (m_curlErrorBuffer[0] != '\0')
					OS::Report("Curl ERROR for url='%s', error: '%s'\n", url, m_curlErrorBuffer);
				else
					OS::Report("Curl ERROR for url='%s', error: '%s'\n", url, curl_easy_strerror(msg->data.result));
				break;

			default:
				OS::Report("Curl ERROR for url='%s', error: '%s'\n", url, curl_easy_strerror(msg->data.result));
			}
#else
            fprintf(stderr, "Curl ERROR for url='%s', error: '%s'\n", url, curl_easy_strerror(msg->data.result));
#endif
#endif
            if (d->client()) {
				if(msg->data.result == CURLE_SSL_CACERT && m_curlErrorBuffer[0] != '\0') {
					d->client()->didFail(job, ResourceError(String(), msg->data.result, String(url), String(m_curlErrorBuffer)));
				} else {
					d->client()->didFail(job, ResourceError(String(), msg->data.result, String(url), String(curl_easy_strerror(msg->data.result))));
				}
			}
        }

        removeFromCurl(job);
    }
    LOG_SIMPLE_RENDER_MSG("    Trues whiled.\n");

    bool started = startScheduledJobs(); // new jobs might have been added in the meantime

	//RJD CRANK:  Due to the slow performance of SSL downloads, it was decided that it would be good to get back in to the 
	// download callback as soon as possible.  After 128 KB has been downloaded, the timer is set to 20 ms to give the main 
	// event loop some time to catch up.  The check for the 128 KB is performed in the write callback above.
#if INCREASE_DOWNLOAD_TIMER
    if (!m_downloadTimer.isActive() && (started || (runningHandles > 0))) {
        if (m_increaseDownloadFrequency) {
			m_downloadTimer.startOneShot(0);
		} else {
			m_downloadTimer.startOneShot(pollTimeSeconds);
			increaseDownloadFrequency(); 
		}
	}
#else
    if (!m_downloadTimer.isActive() && (started || (runningHandles > 0)))
		m_downloadTimer.startOneShot(pollTimeSeconds);
#endif

    LOG_SIMPLE_RENDER_MSG("    EVERYTHING DONE.\n");
}

void ResourceHandleManager::setProxyInfo(const String& host,
                                         unsigned long port,
                                         ProxyType type,
                                         const String& username,
                                         const String& password)
{
    m_proxyType = type;

    if (!host.length()) {
        m_proxy = String("");
    } else {
        String userPass;
        if (username.length() || password.length())
            userPass = username + ":" + password + "@";

        m_proxy = String("http://") + userPass + host + ":" + String::number(port);
    }
}

void ResourceHandleManager::setCertificate(const CString& cert)
{
    m_certificatePath = cert;
}

void ResourceHandleManager::setCrl(const CString& crl)
{
    m_crlPath = crl;
}

void ResourceHandleManager::setClientCertificate(const CString& cert)
{
    m_clientCertificatePath = cert;
}

void ResourceHandleManager::addGlobalCustomHeader(const String& name, const String& value)
{
	// if this header already exists, delete it first
	if ((m_globalCustomHeaders.get(name.latin1().data())).latin1().data()[0] != '\0')
		m_globalCustomHeaders.remove(name.latin1().data());

	m_globalCustomHeaders.add(name, value);
}

void ResourceHandleManager::removeGlobalCustomHeader(const String& name)
{
	m_globalCustomHeaders.remove(name.latin1().data());
}

void ResourceHandleManager::removeAllGlobalCustomHeaders()
{
	m_globalCustomHeaders.clear();
}

void ResourceHandleManager::addAuthenticationMapping(const String& url, const String& user, const String& pass)
{
	m_AuthenticationMap.remove(url);
	m_AuthenticationMap.add(url, pair<String, String>(user, pass));
}

void ResourceHandleManager::removeFromCurl(ResourceHandle* job)
{
    ResourceHandleInternal* d = job->getInternal();
    ASSERT(d->m_handle);
    if (!d->m_handle)
        return;
    m_runningJobs--;
    int size = m_activeHandleList.size();
    for (int i = 0; i < size; i++) {
        if (job == m_activeHandleList[i]) {
            m_activeHandleList.remove(i);
			size = m_activeHandleList.size();
        }
    }


#if  PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
    // RJD Crank:  On the last job, check if the cookie timer is disabled.  If it is, arm it to flush the cookies to disk in three minutes.
    if(size == 0) {
		if (m_cookieLimit != 0 && !m_cookieFlushTimer.isActive()) {
			m_cookieFlushTimer.startOneShot(180);
		}
    }
#endif

    curl_multi_remove_handle(m_curlMultiHandle, d->m_handle);
    curl_easy_cleanup(d->m_handle);
    d->m_handle = 0;
    job->deref();

}

void ResourceHandleManager::setupPUT(ResourceHandle*, struct curl_slist**)
{
    notImplemented();
}

/* Calculate the length of the POST.
   Force chunked data transfer if size of files can't be obtained.
 */
void ResourceHandleManager::setupPOST(ResourceHandle* job, struct curl_slist** headers)
{
    ResourceHandleInternal* d = job->getInternal();
    curl_easy_setopt(d->m_handle, CURLOPT_POST, TRUE);
    curl_easy_setopt(d->m_handle, CURLOPT_POSTFIELDSIZE, 0);

    if (!job->firstRequest().httpBody())
        return;

    Vector<FormDataElement> elements = job->firstRequest().httpBody()->elements();
    size_t numElements = elements.size();
    if (!numElements)
        return;

    // Do not stream for simple POST data
    if (numElements == 1) {
        job->firstRequest().httpBody()->flatten(d->m_postBytes);
        if (d->m_postBytes.size() != 0) {
            curl_easy_setopt(d->m_handle, CURLOPT_POSTFIELDSIZE, d->m_postBytes.size());
            curl_easy_setopt(d->m_handle, CURLOPT_POSTFIELDS, d->m_postBytes.data());
        }
        return;
    }

    // Obtain the total size of the POST
    // The size of a curl_off_t could be different in WebKit and in cURL depending on
    // compilation flags of both. For CURLOPT_POSTFIELDSIZE_LARGE we have to pass the
    // right size or random data will be used as the size.
    static unsigned long expectedSizeOfCurlOffT = 0;

    if (!expectedSizeOfCurlOffT) {
	    ADD_TO_STATIC_LOCAL(expectedSizeOfCurlOffT);
        curl_version_info_data *infoData = curl_version_info(CURLVERSION_NOW);
        if (infoData->features & CURL_VERSION_LARGEFILE)
            expectedSizeOfCurlOffT = sizeof(long long);
        else
            expectedSizeOfCurlOffT = sizeof(int);
    }

#if COMPILER(MSVC)
    // work around compiler error in Visual Studio 2005.  It can't properly
    // handle math with 64-bit constant declarations.
#pragma warning(disable: 4307)
#endif
    static const long long maxCurlOffT = (1LL << (expectedSizeOfCurlOffT * 8 - 1)) - 1;
    curl_off_t size = 0;
    bool chunkedTransfer = false;
    for (size_t i = 0; i < numElements; i++) {
        FormDataElement element = elements[i];
        if (element.m_type == FormDataElement::encodedFile) {
            long long fileSizeResult;
            if (getFileSize(element.m_filename, fileSizeResult)) {
                if (fileSizeResult > maxCurlOffT) {
                    // File size is too big for specifying it to cURL
                    chunkedTransfer = true;
                    break;
                }
                size += fileSizeResult;
            } else {
                chunkedTransfer = true;
                break;
            }
        } else
            size += elements[i].m_data.size();
    }

    // cURL guesses that we want chunked encoding as long as we specify the header
    if (chunkedTransfer)
        *headers = curl_slist_append(*headers, "Transfer-Encoding: chunked");
    else {
        if (sizeof(long long) == expectedSizeOfCurlOffT)
          curl_easy_setopt(d->m_handle, CURLOPT_POSTFIELDSIZE_LARGE, (long long)size);
        else
          curl_easy_setopt(d->m_handle, CURLOPT_POSTFIELDSIZE_LARGE, (int)size);
    }

    curl_easy_setopt(d->m_handle, CURLOPT_READFUNCTION, readCallback);
    curl_easy_setopt(d->m_handle, CURLOPT_READDATA, job);
}

void ResourceHandleManager::add(ResourceHandle* job)
{
#if  PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
    KURL kurl = job->firstRequest().url();
    HTTPCache *cache = 0;
    if (kurl.protocol() == "https") {
        cache = &m_httpsCache; 
    } else {
        cache = &m_httpCache; 
    }

    if (!cache->disabled() && job->firstRequest().cachePolicy() != ReloadIgnoringCacheData) {
        HTTPCachedResource *resource = cache->resourceForURL(kurl);
        if (resource) {
            double override = cache->cacheOverrideFor(resource);

            if (!resource->needRevalidate(override)) {
                // read from the cached resource
                scheduleLoadResourceFromHTTPCache(job);
                return;
            }
            // get from a network unless the cached resource is up-to-date.
            const String& lastModified = resource->lastModifiedHeader();
            if (!lastModified.isEmpty()) {
                job->firstRequest().setHTTPHeaderField("If-Modified-Since", lastModified);
            }
        }
    }
#endif

    // we can be called from within curl, so to avoid re-entrancy issues
    // schedule this job to be added the next time we enter curl download loop
    job->ref();
    m_resourceHandleList.append(job);
    if (!m_downloadTimer.isActive())
        m_downloadTimer.startOneShot(pollTimeSeconds);
}

bool ResourceHandleManager::removeScheduledJob(ResourceHandle* job)
{
    int size = m_resourceHandleList.size();
    for (int i = 0; i < size; i++) {
        if (job == m_resourceHandleList[i]) {
            m_resourceHandleList.remove(i);
            job->deref();
            return true;
        }
    }
    return false;
}

bool ResourceHandleManager::startScheduledJobs()
{
    // TODO: Create a separate stack of jobs for each domain.

    bool started = false;
    while (!m_resourceHandleList.isEmpty() && m_runningJobs < maxRunningJobs) {
        ResourceHandle* job = m_resourceHandleList[0];
        m_resourceHandleList.remove(0);
        startJob(job);
        started = true;
    }
    return started;
}

void ResourceHandleManager::dispatchSynchronousJob(ResourceHandle* job)
{
    KURL kurl = job->firstRequest().url();

    if (kurl.protocolIsData()) {
        handleDataURL(job);
        return;
    }

    ResourceHandleInternal* handle = job->getInternal();

#if LIBCURL_VERSION_NUM > 0x071200
    // If defersLoading is true and we call curl_easy_perform
    // on a paused handle, libcURL would do the transfert anyway
    // and we would assert so force defersLoading to be false.
    handle->m_defersLoading = false;
#endif

    initializeHandle(job);

    // curl_easy_perform blocks until the transfert is finished.
    CURLcode ret =  curl_easy_perform(handle->m_handle);

    if (ret != 0) {
        ResourceError error(String(handle->m_url), ret, String(handle->m_url), String(curl_easy_strerror(ret)));
        handle->client()->didFail(job, error);
    }

    curl_easy_cleanup(handle->m_handle);
}

void ResourceHandleManager::startJob(ResourceHandle* job)
{
    KURL kurl = job->firstRequest().url();

    if (kurl.protocolIsData()) {
        handleDataURL(job);
        return;
    }

#if PLATFORM(WIIU)
	int error = WebCore::DomainAccessControl::DOMAIN_ACCESS_OK;
	if (!WebCore::DomainAccessControl::IsDomainAccessible(kurl.string(), &error))
	{
		char szError[2048] = {0};
		switch (error)
		{
		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NO_NETWORK_ACCESS:
			snprintf(szError, sizeof(szError), "This request cannot be honored. You must enable network access.\n\tRequested URL: %s.\n",
				encodeWithURLEscapeSequences(kurl.string()).latin1().data());

			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ASSET_LOAD_FAIL, L"", 0);
			break;

		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ENCRYPTED:
			snprintf(szError, sizeof(szError), "This request must be over HTTPS. Please use the HTTPS protocol.\n\tRequested URL: %s.\n",
				encodeWithURLEscapeSequences(kurl.string()).latin1().data());

			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_CURLE_UNSUPPORTED_PROTOCOL, L"", 0);
			break;

		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_WHITELISTED:
			snprintf(szError, sizeof(szError), "The domain '%s' is not whitelisted. Add it to the Whitelist page of Project Settings.\n",
					 encodeWithURLEscapeSequences(kurl.string()).latin1().data());

			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ADDRESS_BLOCKED_BY_WHITELIST, L"", 0);
			break;

		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN:
			snprintf(szError, sizeof(szError), "Before accessing a network resource, an independent service token must first be acquired.\n" \
					"Please call nwf.act.NintendoAccountManager.acquireNewIndependentServiceToken().\n");
			break;

		default:
			snprintf(szError, sizeof(szError), "This domain is unreachable. Error %d\n", error);
			LOG(Network, "%s", szError);
			break;
		}

		OS::Report("%s", szError);
#ifndef NDEBUG
		fprintf(stderr, "%s", szError);
#endif

        job->cancel();
		return;
	}
#endif

    initializeHandle(job);

    m_activeHandleList.append(job);
    m_runningJobs++;
    CURLMcode ret = curl_multi_add_handle(m_curlMultiHandle, job->getInternal()->m_handle);
    // don't call perform, because events must be async
    // timeout will occur and do curl_multi_perform
    if (ret && ret != CURLM_CALL_MULTI_PERFORM) {
#ifndef NDEBUG
#if PLATFORM(WIIU)
    	OS::Report("Error %d starting job %s\n", ret, encodeWithURLEscapeSequences(job->firstRequest().url().string()).latin1().data());
#else
        fprintf(stderr, "Error %d starting job %s\n", ret, encodeWithURLEscapeSequences(job->firstRequest().url().string()).latin1().data());
#endif
#endif
        job->cancel();
        return;
    }
}

#ifdef _NWF_SET_SOCKET_BUFFER_
static int _StaticSocketCallback(void *clientp, curl_socket_t curlfd, curlsocktype purpose)
	{
	int val = 0;
	int rval = 0;
	
	// Enable TCP selective acknowledgement (SACK)
	val = 1;
	rval = setsockopt(curlfd, SOL_SOCKET, SO_TCPSACK, &val, sizeof(val));
    if( rval != 0 )
    {
        OS::Report("setsockopt on SO_TCPSACK failed with error: %d\n\n", rval);
        return 1;
    }
	
	// Enable TCP window scaling
	val = 1;
	rval = setsockopt(curlfd, SOL_SOCKET, 0x400, &val, sizeof(val));
    if( rval != 0 )
    {
        OS::Report("setsockopt on SO_WINSCALE failed with error: %d\n\n", rval);
        return 1;
    }

	val = 64 * 1024 - 1;
	rval = setsockopt(curlfd, SOL_SOCKET, SO_RCVBUF, &val, sizeof(val));
    if( rval != 0 )
    {
        OS::Report("setsockopt on SO_RCVBUF failed with error: %d\n\n", rval);
        return 1;
    }

	return 0;
	}
#endif

static bool isInvalidHeaderNameCharacter(UChar c)
{
	// according to RFC 2616, section 4.2, HTTP header names do not allow:
	//		control characters (< 0x20 && 0x7f) and space (0x20)
	// or the following separaters: ( ) < > @ , ; : \ " / [ ] ? = { }

    return c <= 0x20 || c == 0x7F
		|| c == '(' || c == ')' || c == '<' || c == '>'
		|| c == '@' || c == ',' || c == ';' || c == ':'
		|| c == '\\' || c == '\"' || c == '/' || c == '['
		|| c == ']' || c == '?' || c == '=' || c == '{' || c == '}';
}

static bool isInvalidHeaderValueCharacter(UChar c)
{
	// according to RFC 2616, section 4.2, HTTP header values do not allow:
	//		control characters (< 0x20 && 0x7f) except horizontal tab (0x09)

    return (c < 0x20 && c != 0x09) || c == 0x7F;
}

String CleanUpHTTPHeaderName(const String &headerName)
{
	return headerName.removeCharacters(&isInvalidHeaderNameCharacter);
}

String CleanUpHTTPHeaderValue(const String &headerValue)
{
	return headerValue.stripWhiteSpace().removeCharacters(&isInvalidHeaderValueCharacter);
}

void ResourceHandleManager::initializeHandle(ResourceHandle* job)
{
	// Don't start a new job if we are in the middle of a reload; it will cause a crash.
	extern bool g_AllowScriptExecution;
	if (!g_AllowScriptExecution)
	{
		job->cancel();
		return;
	}

    KURL kurl = job->firstRequest().url();

    // Remove any fragment part, otherwise curl will send it as part of the request.
    kurl.removeFragmentIdentifier();

    ResourceHandleInternal* d = job->getInternal();
    String url = kurl.string();

    if (kurl.isLocalFile()) {
        String query = kurl.query();
        // Remove any query part sent to a local file.
        if (!query.isEmpty()) {
            int queryIndex = url.find(query);
            if (queryIndex != -1)
                url = url.left(queryIndex - 1);
        }
        // Determine the MIME type based on the path.
        d->m_response.setMimeType(MIMETypeRegistry::getMIMETypeForPath(url));
    }

    d->m_handle = curl_easy_init();

	if (!d->m_handle)
	{
		OS::Report("curl_easy_init failed connect to %s\n", url.latin1().data());
		cancel(job);
		return;
	}

#if LIBCURL_VERSION_NUM > 0x071200
    if (d->m_defersLoading) {
        CURLcode error = curl_easy_pause(d->m_handle, CURLPAUSE_ALL);
        // If we did not pause the handle, we would ASSERT in the
        // header callback. So just assert here.
        ASSERT_UNUSED(error, error == CURLE_OK);
    }
#endif
#ifndef NDEBUG
    if (getenv("DEBUG_CURL"))
        curl_easy_setopt(d->m_handle, CURLOPT_VERBOSE, 1);
#endif
    curl_easy_setopt(d->m_handle, CURLOPT_PRIVATE, job);
    curl_easy_setopt(d->m_handle, CURLOPT_ERRORBUFFER, m_curlErrorBuffer);
    curl_easy_setopt(d->m_handle, CURLOPT_WRITEFUNCTION, writeCallback);
    curl_easy_setopt(d->m_handle, CURLOPT_WRITEDATA, job);
    curl_easy_setopt(d->m_handle, CURLOPT_HEADERFUNCTION, headerCallback);
    curl_easy_setopt(d->m_handle, CURLOPT_WRITEHEADER, job);
    curl_easy_setopt(d->m_handle, CURLOPT_AUTOREFERER, 1);
    curl_easy_setopt(d->m_handle, CURLOPT_FOLLOWLOCATION, 1);
    curl_easy_setopt(d->m_handle, CURLOPT_HTTPAUTH, CURLAUTH_ANY);
    curl_easy_setopt(d->m_handle, CURLOPT_SHARE, m_curlShareHandle);
    curl_easy_setopt(d->m_handle, CURLOPT_DNS_CACHE_TIMEOUT, 60 * 5); // 5 minutes
    curl_easy_setopt(d->m_handle, CURLOPT_SSL_CIPHER_LIST, LibPlatform::SSL::GetCipherList());
	if (ignoreSSLErrors)
        curl_easy_setopt(d->m_handle, CURLOPT_SSL_VERIFYPEER, false);
	
    if (!m_certificatePath.isNull() && m_certificatePath.data()[0] != '\0') {
		curl_easy_setopt(d->m_handle, CURLOPT_CAINFO, m_certificatePath.data());
	}

	if (!m_crlPath.isNull() && m_crlPath.data()[0] != '\0') {
		curl_easy_setopt(d->m_handle, CURLOPT_CRLFILE, m_crlPath.data());
	}

	if (!m_clientCertificatePath.isNull() && m_clientCertificatePath.data()[0] != '\0') {
		curl_easy_setopt(d->m_handle, CURLOPT_SSLCERT, m_clientCertificatePath.data());
	}

    // enable gzip and deflate through Accept-Encoding:
    curl_easy_setopt(d->m_handle, CURLOPT_ENCODING, "");

    // url must remain valid through the request
    ASSERT(!d->m_url);

    // url is in ASCII so latin1() will only convert it to char* without character translation.
    d->m_url = fastStrDup(url.latin1().data());
    curl_easy_setopt(d->m_handle, CURLOPT_URL, d->m_url);

    if (m_cookieLimit != 0 && m_cookieJarFileName && m_loadedCookies == false) {
    	curl_easy_setopt(d->m_handle, CURLOPT_COOKIEFILE, m_cookieJarFileName);
        curl_easy_setopt(d->m_handle, CURLOPT_COOKIEJAR, m_cookieJarFileName);
        m_loadedCookies = true;
    }

#ifdef _NWF_SET_SOCKET_BUFFER_
	curl_easy_setopt(d->m_handle, CURLOPT_SOCKOPTFUNCTION, _StaticSocketCallback);
#endif

    struct curl_slist* headers = 0;
    if (job->firstRequest().httpHeaderFields().size() > 0) {
        HTTPHeaderMap customHeaders = job->firstRequest().httpHeaderFields();
        HTTPHeaderMap::const_iterator end = customHeaders.end();
        for (HTTPHeaderMap::const_iterator it = customHeaders.begin(); it != end; ++it) {
            String key = it->first;
			String value = it->second;
            String headerString(key);
            headerString.append(": ");
            headerString.append(value);
            CString headerLatin1 = headerString.latin1();
            headers = curl_slist_append(headers, headerLatin1.data());
        }
    }

	if(WebViewHandle_t view = WebViewTV())
	{
		std::vector<String> ch;
		if (((WebView *)view)->GetCustomHeaders(&ch))
		{
			for (unsigned int HeaderIndex = 0; HeaderIndex < ch.size(); HeaderIndex++)
			{
				// parse the header before calling add
				Vector<String> header;
				ch[HeaderIndex].split(":", header);
				if (header.size() == 2)
					m_globalCustomHeaders.add(CleanUpHTTPHeaderName(header[0]), CleanUpHTTPHeaderValue(header[1]));
			}
		}
	}

	if (m_globalCustomHeaders.size() > 0) {
        HTTPHeaderMap::const_iterator end = m_globalCustomHeaders.end();
        for (HTTPHeaderMap::const_iterator it = m_globalCustomHeaders.begin(); it != end; ++it) {
            String key = it->first;
            String value = it->second;
            String headerString(key);
            headerString.append(": ");
            headerString.append(value);
            CString headerLatin1 = headerString.latin1();
            headers = curl_slist_append(headers, headerLatin1.data());
        }
    }

    if ("GET" == job->firstRequest().httpMethod())
        curl_easy_setopt(d->m_handle, CURLOPT_HTTPGET, TRUE);
    else if ("POST" == job->firstRequest().httpMethod())
        setupPOST(job, &headers);
    else if ("PUT" == job->firstRequest().httpMethod())
        setupPUT(job, &headers);
    else if ("HEAD" == job->firstRequest().httpMethod())
        curl_easy_setopt(d->m_handle, CURLOPT_NOBODY, TRUE);
	else if ("OPTIONS" == job->firstRequest().httpMethod())
		curl_easy_setopt(d->m_handle, CURLOPT_CUSTOMREQUEST, "OPTIONS");

    if (headers) {
        curl_easy_setopt(d->m_handle, CURLOPT_HTTPHEADER, headers);
        d->m_customHeaders = headers;
    }

    HashMap<String, pair<String, String> >::const_iterator end = m_AuthenticationMap.end();
    for (HashMap<String, pair<String, String> >::const_iterator it = m_AuthenticationMap.begin(); it != end; ++it) {
        bool bAddMapping = false;

        if ((*it).first.contains('*')) {
            String strAuthenticate = (*it).first;
            if (strAuthenticate == "*")
                bAddMapping = true;
            else {
                int idx = strAuthenticate.find("://");
                String strProtocol = strAuthenticate.substring(0, idx);
                if (strProtocol == kurl.protocol()) {
                    String strHost = strAuthenticate.substring(idx + 3);
                    if (strHost.startsWith("*")) {
                        String strHostPostWildcard = strHost.substring(1);
                        if (url.contains(strHostPostWildcard))
                            bAddMapping = true;
                    }
                }
            }
        } else {
            KURL kurlAuthenticate(ParsedURLString, (*it).first);    // let KURL parse it first so it'll pretty it up
            String strAuthenticate = kurlAuthenticate.string();
            int idx = strAuthenticate.find("://");
            String strProtocol = strAuthenticate.substring(0, idx);
            if (strProtocol == kurl.protocol()) {
                String strHost = strAuthenticate.substring(kurlAuthenticate.hostStart(), kurlAuthenticate.hostEnd() - kurlAuthenticate.hostStart());
                String requestHost = url.substring(kurl.hostStart(), kurl.hostEnd() - kurl.hostStart());
                if (requestHost == strHost)
                    bAddMapping = true;
            }
        }

        if (bAddMapping) {
            d->m_user = (*it).second.first;
            d->m_pass = (*it).second.second;
        }
    }

    // curl CURLOPT_USERPWD expects username:password
    if (d->m_user.length() || d->m_pass.length()) {
        String userpass = d->m_user + ":" + d->m_pass;
        curl_easy_setopt(d->m_handle, CURLOPT_USERPWD, userpass.utf8().data());
    }

    // Set proxy options if we have them.
    if (m_proxy.length()) {
        curl_easy_setopt(d->m_handle, CURLOPT_PROXY, m_proxy.utf8().data());
        curl_easy_setopt(d->m_handle, CURLOPT_PROXYTYPE, m_proxyType);
    }
}

void ResourceHandleManager::cancel(ResourceHandle* job)
{
    if (removeScheduledJob(job))
        return;

    ResourceHandleInternal* d = job->getInternal();
    d->m_cancelled = true;
    if (!m_downloadTimer.isActive())
        m_downloadTimer.startOneShot(pollTimeSeconds);
}

#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
void ResourceHandleManager::cancel(const String& strURL)
{
	KURL kurl(KURL(), strURL);
	Vector<ResourceHandle *>::const_iterator end = m_activeHandleList.end();
    for (Vector<ResourceHandle *>::const_iterator it = m_activeHandleList.begin(); it != end; ++it) {
        ResourceHandle *rh = *it;
        if (rh->firstRequest().url().string() == kurl.string()) {
            cancel(rh);
            break;
        }
    }
}

void ResourceHandleManager::cancelAuthentication(const String& strURL)
{
	KURL kurl(KURL(), strURL);
	Vector<ResourceHandle *>::const_iterator end = m_activeHandleList.end();
    for (Vector<ResourceHandle *>::const_iterator it = m_activeHandleList.begin(); it != end; ++it) {
        ResourceHandle *rh = *it;
        if (rh->firstRequest().url().string() == kurl.string()) {
            ResourceHandleInternal* d = rh->getInternal();
			d->m_cancelAuthentication = true;
            break;
        }
    }
}

void ResourceHandleManager::flushCookiesCallback(Timer<ResourceHandleManager>* timer) {
	ThreadIdentifier id = createThread(ResourceHandleManager::flushCookiesThread, this, "Cookie Thread");
	detachThread(id);
}

/*!
 	Flush the cookies to disk
 */

void *ResourceHandleManager::flushCookiesThread(void *data) {
	ResourceHandleManager::sharedInstance()->flushCookiesToDisk();
	return 0;
}

/*!
 	Clear all of the current session cookies that are stored.
 */
void ResourceHandleManager::clearSessionCookies() {
	CURL *curl = curl_easy_init();
	if (!curl) {
		return;
	}

	curl_easy_setopt(curl, CURLOPT_SHARE, m_curlShareHandle);
	if (m_cookieJarFileName) {
		curl_easy_setopt(curl, CURLOPT_COOKIEJAR, m_cookieJarFileName);
	}

	curl_easy_setopt(curl, CURLOPT_COOKIELIST, "SESS");
	curl_easy_cleanup(curl);
}

/*!
 	Clear all of the current cookies that are stored.  This will remove the cookie file as well.
 */
void ResourceHandleManager::clearAllCookies() {
	CURL *curl = curl_easy_init();
	if (!curl) {
		return;
	}
	 
	curl_easy_setopt(curl, CURLOPT_SHARE, m_curlShareHandle);
	if (m_cookieJarFileName) {
		curl_easy_setopt(curl, CURLOPT_COOKIEJAR, m_cookieJarFileName);
	}
	
	curl_easy_setopt(curl, CURLOPT_COOKIELIST, "ALL");
	curl_easy_cleanup(curl);
	unlink(m_cookieJarFileName);
}

/*!
 	Save all of the current persistent cookies to disk.
 */
void ResourceHandleManager::flushCookiesToDisk() {
	long cookie_size = 0;

	MutexLocker locker(m_flushCookiesLock);
	//LibPlatform::OS::Report("[flushCookiesToDisk]Flushing to disk\n"); 
	CURL *curl = curl_easy_init();
	if (!curl) {
		return;
	}
	curl_easy_setopt(curl, CURLOPT_SHARE, m_curlShareHandle);
	if (m_cookieJarFileName) {
		curl_easy_setopt(curl, CURLOPT_COOKIEJAR, m_cookieJarFileName);
	}


	curl_easy_getinfo(curl, CURLINFO_COOKIE_SIZE, &cookie_size);
	//LibPlatform::OS::Report("[flushCookiesToDisk]Cookie Limit %ld\n", m_cookieLimit);
	if(m_cookieLimit != -1 && cookie_size > m_cookieLimit) {
		//LibPlatform::OS::Report("[flushCookiesToDisk]Cookie size on disk %ld greater than limit %ld\n", cookie_size, m_cookieLimit);
		LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_CURLE_WRITE_ERROR, L"Cookie size exceeds file limit", 0);
		curl_easy_setopt(curl, CURLOPT_COOKIEFILE, NULL);
		curl_easy_setopt(curl, CURLOPT_COOKIEJAR, NULL);
		curl_easy_cleanup(curl);
		return;
	}

	curl_easy_setopt(curl, CURLOPT_COOKIELIST, "FLUSH");
	curl_easy_setopt(curl, CURLOPT_COOKIEFILE, NULL);
	curl_easy_setopt(curl, CURLOPT_COOKIEJAR, NULL);
	curl_easy_cleanup(curl);
	//LibPlatform::OS::Report("[flushCookiesToDisk]Flush complete\n"); 
}

HTTPCachedResource* ResourceHandleManager::updateCacheResource(KURL &url, RefPtr<SharedBuffer> resourceData, ResourceResponse &response, bool noCache, bool noStore, bool mustRevalidate, double expires, double maxAge)
{
    HTTPCache *cache = 0;
    if (url.protocol() == "https") {
        cache = &m_httpsCache; 
    } else {
        cache = &m_httpCache; 
    }

    HTTPCachedResource *resource = cache->resourceForURL(url);
    if (resource) {
        if (noStore) {
            cache->remove(resource);
            return NULL;
        } else {
            cache->updateCachedResource(resource, resourceData, response, noCache, mustRevalidate, expires, maxAge);
            cache->detach(resource);
            return resource;
        }
    }
    
    for(int num = 0; num < m_writeCacheList.size(); num++) {
        if (cache->equalHTTPCachedResourceURL(m_writeCacheList[num], url)) {
            if (noStore) {
                m_writeCacheList.remove(num);
                return NULL;
            }
            resource = m_writeCacheList[num];
            cache->updateCachedResource(resource, resourceData, response, noCache, mustRevalidate, expires, maxAge);
            m_writeCacheList.remove(num);
            break;
        }
    }
    return resource;
}


bool ResourceHandleManager::addHTTPCache(ResourceHandle *handle, KURL &url, RefPtr<SharedBuffer> resourceData, ResourceResponse &response)
{
	MutexLocker locker(m_cacheAccessLock);
    if (!resourceData)
        return false;

    if (!url.protocolInHTTPFamily())
        return false;

    bool noCache = response.cacheControlContainsNoCache();
    bool noStore = response.cacheControlContainsNoStore();
    bool mustRevalidate = response.cacheControlContainsMustRevalidate();
    double expires = std::numeric_limits<double>::quiet_NaN();
    double maxAge = std::numeric_limits<double>::quiet_NaN();

    HTTPCache *cache = 0;
    if (url.protocol() == "https") {
        cache = &m_httpsCache; 
    } else {
        cache = &m_httpCache; 
    }

    HTTPCachedResource *resource = updateCacheResource(url, resourceData, response, noCache, noStore, mustRevalidate, expires, maxAge);
    if (!resource && !noStore) {
        resource = cache->createHTTPCachedResource(url, resourceData, response, noCache, mustRevalidate, expires, maxAge);

        if (!resource) {
            return false;
        }
    }
    if (noStore) {
        return false;
    }

    m_writeCacheList.append(resource);

    if (!m_writeCacheTimer.isActive()) {
        //use main cache for interval
        m_writeCacheTimer.startOneShot(cache->getWriteInterval());
    }

    return true;
}

void ResourceHandleManager::scheduleLoadResourceFromHTTPCache(ResourceHandle *job)
{
    job->ref();
    m_readCacheJobList.append(job);
    if (!m_readCacheTimer.isActive()) {
        m_readCacheTimer.startOneShot(pollTimeSeconds);
    }
}

void ResourceHandleManager::readCacheTimerCallback(Timer<ResourceHandleManager>* timer)
{
    ResourceHandle* job;
	MutexLocker locker(m_cacheAccessLock);
    if (m_readCacheJobList.size() == 0)
        return;

    job = m_readCacheJobList.first();
    KURL kurl = job->firstRequest().url();

    HTTPCache *cache = 0;
    if (kurl.protocol() == "https") {
        cache = &m_httpsCache; 
    } else {
        cache = &m_httpCache; 
    }

    HTTPCachedResource *resource = cache->resourceForURL(kurl);

    ResourceHandleInternal* d = job->getInternal();

    if (!d->m_cancelled && d->client() && resource) {

        if (!d->client() || d->m_cancelled)
            goto cancel;

        char *buf;
        buf = (char*)fastMalloc(resource->contentLength());
        if (buf) {
            if (cache->read(resource, buf)) {
                d->m_response.setURL(kurl);
                d->m_response.setMimeType(resource->mimeType());
                d->m_response.setExpectedContentLength(resource->expectedContentLength());
                d->m_response.setTextEncodingName(resource->textEncodingName());
                d->m_response.setSuggestedFilename(resource->suggestedFilename());
                d->m_response.setHTTPStatusCode(resource->httpStatusCode());
                d->m_response.setWasCached(true);
                d->client()->didReceiveResponse(job, d->m_response);

                if (!d->client() || d->m_cancelled)
                    goto cancel;
                d->client()->didReceiveData(job, buf, resource->contentLength(), 0);

                if (!d->client() || d->m_cancelled)
                    goto cancel;
                d->client()->didFinishLoading(job, currentTime());
            } else {
                //Resource failed to load, remove resource from cache and restart the job
                cache->removeResource(resource);

                if (d->m_response.httpStatusCode() == 304) {
                    free(d->m_url);
                    free(d->m_customHeaders);
                    d->m_url = NULL;
                    d->m_customHeaders = NULL;
                    d->m_response = ResourceResponse();
                    job->firstRequest().setHTTPHeaderField("If-Modified-Since", "");
                }
                add(job);
            }
            fastFree(buf);
        }
    }

    m_readCacheJobList.remove(0);
    job->deref();

    if (m_readCacheJobList.size() == 0)
        return;

    if (!m_readCacheTimer.isActive())
        m_readCacheTimer.startOneShot(pollTimeSeconds);

    return;

cancel:
    if (!d->m_cancelled) {
        cancel(job);
        if (d->client())
            d->client()->didFail(job, ResourceError(String(), CURLE_READ_ERROR, String(d->m_url), String("cache read error")));
    }

    if (m_readCacheJobList.size() == 0)
        return;

    if (!m_readCacheTimer.isActive())
        m_readCacheTimer.startOneShot(pollTimeSeconds);
}

void ResourceHandleManager::writeCacheTimerCallback(Timer<ResourceHandleManager>* timer)
{
	ThreadIdentifier id = createThread(ResourceHandleManager::writeCacheThread, this, "File Cache Write Thread");
	detachThread(id);
}

void *ResourceHandleManager::writeCacheThread(void *data)
{
	ResourceHandleManager::sharedInstance()->writeCache();
	return 0;
}

void ResourceHandleManager::writeCache()
{
	// Make sure that only one writeCache() call is performed at a time.
	MutexLocker threadLocker(m_writeCacheLock);
	MutexLocker cachelocker(m_cacheAccessLock);
	if (m_writeCacheList.size() == 0)
		return;

	if (!m_resourceHandleList.isEmpty()) {
		//use main cache for interval
		callOnMainThread(cleanupCacheResourceData, 0);
		return;
	}

	//reserve enough space for FAT file
	size_t totalWritten = m_httpCache.getFATFileSize() + m_httpsCache.getFATFileSize();

	//we use the main cache to get the write limit
	while (totalWritten < m_httpCache.getWriteLimit() && m_writeCacheList.size()) {

		HTTPCachedResource *resource = m_writeCacheList.last();
		size_t fileSize = resource->getFileSize();

		if (totalWritten + fileSize > m_httpCache.getWriteLimit()) {
			break;
		}

		bool write;
		if (resource->url().startsWith("https")) {
			write = m_httpsCache.write(resource);
		} else {
			write = m_httpCache.write(resource);
		}

		size_t size = m_writeCacheList.size();
		m_writeCacheList.remove(size - 1);
		if (!write) {
			m_cachedResourceDeletionList.append(resource);
		} else {
			totalWritten += fileSize;
			//OS::Report("written %d of %d\n", totalWritten, m_httpCache.getWriteLimit());
		}
	}

	//Crank ML: Write the fat files before deleting any cache files.
	m_httpCache.writeFATFile();
	m_httpsCache.writeFATFile();

	//remove all the cache that wasn't able to be written
	while (m_writeCacheList.size() > 0)
	{
		HTTPCachedResource *resource = m_writeCacheList.first();
		m_cachedResourceDeletionList.append(resource);
		m_writeCacheList.remove(0);
	}

	callOnMainThread(cleanupCacheResourceData, 0);


}

void ResourceHandleManager::forceTerminateHTTPCache()
{
	MutexLocker locker(m_cacheAccessLock);
    ResourceHandle* job;
    int num = m_readCacheJobList.size();

    m_readCacheTimer.stop();
    while (num--) {
        job = m_readCacheJobList[num];
        m_readCacheJobList.remove(num);
        job->deref();
    }

    m_writeCacheTimer.stop();

    while (m_writeCacheList.size()) {

        HTTPCachedResource *resource = m_writeCacheList.first();

        bool write = m_httpCache.write(resource);
        m_writeCacheList.remove(0);
        if (!write) {
            delete resource;
        }
    }

    m_httpCache.writeFATFile(); 
    m_httpCache.forceTerminate();

    m_httpsCache.writeFATFile();
    m_httpsCache.forceTerminate();
}

void ResourceHandleManager::parseCookieConfig() {
	xmlDocPtr doc;
    const char *config = shim_GetConfigXMLFile();
    int configSize = shim_GetConfigXMLSize();

    if (configSize) {
        doc = xmlReadMemory(config, configSize, NULL, NULL, 0);
    }
    else {
        doc = xmlReadFile(config, NULL, 0);
    }
    if (doc == NULL) {
        //OS::Report("warning: could not parse cache configuration file: using default values");
        return;
    }

    xmlNode *root = NULL, *cur = NULL;
    root = xmlDocGetRootElement(doc);

    for (cur = root->children; cur; cur = cur->next) {
        if (xmlStrEqual(cur->name, (xmlChar *)"options")) {
            //found options, look for cache config in its children
            cur = cur->children;
            continue;
        }

        if (xmlStrEqual(cur->name, (xmlChar *)"cookieConfig")) {
            //found cookie config, we can stop searching
            root = cur;
            break;
        }
    }

    if (root == xmlDocGetRootElement(doc)) {
        xmlFreeDoc(doc);
        //OS::Report("warning: configuration file has no cache section: using default values");
        return;
    }

    for (cur = root->children; cur; cur = cur->next) {
    	if (cur->type == XML_ELEMENT_NODE) {
    		xmlNode *child = cur->children;

    		//xmlChar is typedef'd to be unsigned char, so we end up
    		//doing a lot of casts back to char below:
    		if (child) {
    			if (xmlStrEqual(cur->name, (xmlChar *)"maxSize")) {
    				if (child->content) {
    					m_cookieLimit = atoi((char *)child->content);
    				}
    			}
    		}
    	}
    }
    xmlFreeDoc(doc);

}

void ResourceHandleManager::cleanupCacheResources()
{
	m_httpCache.cleanUpWrittenResources();
	m_httpsCache.cleanUpWrittenResources();

	m_httpCache.cleanUpRemovedResources();
	m_httpsCache.cleanUpRemovedResources();

	while (m_cachedResourceDeletionList.size() > 0) {
		HTTPCachedResource *resource = m_cachedResourceDeletionList.first();
		delete resource;
		m_cachedResourceDeletionList.remove(0);
	}

	if (!m_resourceHandleList.isEmpty()) {
		m_writeCacheTimer.startOneShot(m_httpCache.getWriteInterval());
	}
}

#endif

} // namespace WebCore
