/*
 *  Copyright (C) 2008 INdT - Instituto Nokia de Tecnologia
 *  Copyright (C) 2009-2010 ProFUSION embedded systems
 *  Copyright (C) 2009-2010 Samsung Electronics
 * Copyright (c) 2012 NINTENDO.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "CachedImage.h"
#include "ClipboardWiiU.h"
#include "DragData.h"
#include "Frame.h"
#include "DataTransferItemsWiiU.h"
#include "Editor.h"
#include "FileList.h"
#include "HTMLNames.h"
#include "HTMLParserIdioms.h"
#include "Image.h"
#include "markup.h"
#include "MIMETypeRegistry.h"
#include "RenderImage.h"
#include "ScriptExecutionContext.h"
#include <wtf/text/StringHash.h>
#include <wtf/OwnPtr.h>

namespace WebCore {


static String normalizeType(const String& type)
{
    String cleanType = type.stripWhiteSpace().lower();
    if (cleanType == mimeTypeText || cleanType.startsWith(mimeTypeTextPlainEtc))
        return mimeTypeTextPlain;
    return cleanType;
}

PassRefPtr<Clipboard> Editor::newGeneralClipboard(ClipboardAccessPolicy policy, Frame* frame)
{
    return ClipboardWiiU::create(Clipboard::CopyAndPaste, WiiUDataObject::create(Clipboard::CopyAndPaste), policy, frame);
}

PassRefPtr<Clipboard> Clipboard::create(ClipboardAccessPolicy policy, DragData* dragData, Frame* frame)
{
    return  ClipboardWiiU::create(DragAndDrop, dragData->platformData(), policy, frame);
}

ClipboardWiiU::ClipboardWiiU(ClipboardType clipboardType,
                                     PassRefPtr<WiiUDataObject> dataObject,
                                     ClipboardAccessPolicy policy,
                                     Frame* frame)
    : Clipboard(policy, clipboardType)
    , m_dataObject(dataObject)
    , m_frame(frame)
    , m_originalSequenceNumber(m_dataObject->getSequenceNumber())
{
}

PassRefPtr<ClipboardWiiU> ClipboardWiiU::create(ClipboardType clipboardType,
    PassRefPtr<WiiUDataObject> dataObject, ClipboardAccessPolicy policy, Frame* frame)
{
    return adoptRef(new ClipboardWiiU(clipboardType, dataObject, policy, frame));
}

ClipboardWiiU::~ClipboardWiiU()
{
}

void ClipboardWiiU::clearData(const String& type)
{
    if (policy() != ClipboardWritable || !m_dataObject)
        return;

    m_dataObject->clearData(normalizeType(type));
}

void ClipboardWiiU::writePlainText(const WTF::String& text)
{
    if (!m_dataObject)
        return;

    String str = text;
    static const UChar NonBreakingSpaceCharacter = 0xA0;
    static const UChar SpaceCharacter = ' ';
    str.replace(NonBreakingSpaceCharacter, SpaceCharacter);
    m_dataObject->setData(mimeTypeTextPlain, str);
}

void ClipboardWiiU::clearAllData()
{
    if (policy() != ClipboardWritable)
        return;

    m_dataObject->clearAll();
}

String ClipboardWiiU::getData(const String& type, bool &success) const
{
    success = false;
    if (policy() != ClipboardReadable || !m_dataObject)
        return String();

    if (m_dataObject->getSequenceNumber() != m_originalSequenceNumber)
        return String();

    return m_dataObject->getData(normalizeType(type), success);
}

bool ClipboardWiiU::setData(const String& type, const String& data)
{
    if (policy() != ClipboardWritable)
        return false;

    return m_dataObject->setData(normalizeType(type), data);
}

HashSet<String> ClipboardWiiU::types() const
{
    HashSet<String> results;
    if (policy() != ClipboardReadable && policy() != ClipboardTypesReadable)
        return results;

    if (!m_dataObject)
        return results;

    results = m_dataObject->types();

    if (m_dataObject->containsFilenames())
        results.add(mimeTypeFiles);

    return results;
}

PassRefPtr<FileList> ClipboardWiiU::files() const
{
    if (policy() != ClipboardReadable)
        return FileList::create();

    if (!m_dataObject)
        return FileList::create();

    const Vector<String>& filenames = m_dataObject->filenames();
    RefPtr<FileList> fileList = FileList::create();
    for (size_t i = 0; i < filenames.size(); ++i)
        fileList->append(File::create(filenames.at(i)));

    return fileList.release();
}

void ClipboardWiiU::setDragImage(CachedImage* image, const IntPoint& loc)
{
    setDragImage(image, 0, loc);
}
void ClipboardWiiU::setDragImage(CachedImage* image, Node *node, const IntPoint &loc)
{
    if (policy() != ClipboardImageWritable && policy() != ClipboardWritable)
        return;

    if (m_dragImage)
        m_dragImage->removeClient(this);
    m_dragImage = image;
    if (m_dragImage)
        m_dragImage->addClient(this);

    m_dragLoc = loc;
    m_dragImageElement = node;
}

void ClipboardWiiU::setDragImageElement(Node* node, const IntPoint& loc)
{
    setDragImage(0, node, loc);
}

DragImageRef ClipboardWiiU::createDragImage(IntPoint& dragLoc) const
{
    if (m_dragImage) {
    	dragLoc = m_dragLoc;
    	return m_dragImage->image()->nativeImageForCurrentFrame();
    } else if (m_dragImageElement) {
    	dragLoc = m_dragLoc;
    	Node* node = m_dragImageElement.get();
    	return node->document()->frame()->nodeImage(node);
    }
    return 0;
}


static CachedImage* getCachedImage(Element* element)
{
    // Attempt to pull CachedImage from element
    ASSERT(element);
    RenderObject* renderer = element->renderer();
    if (!renderer || !renderer->isImage())
        return 0;

    RenderImage* image = toRenderImage(renderer);
    if (image->cachedImage() && !image->cachedImage()->errorOccurred())
        return image->cachedImage();

    return 0;
}

static void writeImageToDataObject(WiiUDataObject* dataObject, Element* element,
                                   const KURL& url)
{
    // Shove image data into a DataObject for use as a file
    CachedImage* cachedImage = getCachedImage(element);
    if (!cachedImage || !cachedImage->image() || !cachedImage->isLoaded())
        return;

    SharedBuffer* imageBuffer = cachedImage->image()->data();
    if (!imageBuffer || !imageBuffer->size())
        return;

    dataObject->setFileContent(imageBuffer);

    // Determine the filename for the file contents of the image.
    String filename = cachedImage->response().suggestedFilename();
    if (filename.isEmpty())
        filename = url.lastPathComponent();
    if (filename.isEmpty())
        filename = element->getAttribute(HTMLNames::altAttr);
    else {
        // Strip any existing extension. Assume that alt text is usually not a filename.
        int extensionIndex = filename.reverseFind('.');
        if (extensionIndex != -1)
            filename.truncate(extensionIndex);
    }

    String extension = MIMETypeRegistry::getMIMETypeForExtension(
        cachedImage->response().mimeType());
    dataObject->setFileExtension(extension.isEmpty() ? emptyString() : "." + extension);

    dataObject->setFileContentFilename(filename + dataObject->fileExtension());
}

void ClipboardWiiU::declareAndWriteDragImage(Element* element, const KURL& url, const String& title, Frame* frame)
{
    if (!m_dataObject)
        return;

    m_dataObject->setData(mimeTypeURL, url);
    m_dataObject->setUrlTitle(title);

    // Write the bytes in the image to the file format.
    writeImageToDataObject(m_dataObject.get(), element, url);

    // Put img tag on the clipboard referencing the image
    m_dataObject->setData(mimeTypeTextHTML, createMarkup(element, IncludeNode, 0, ResolveAllURLs));
}

void ClipboardWiiU::writeURL(const KURL& url, const String& title, Frame*)
{
    if (!m_dataObject)
        return;
    ASSERT(!url.isEmpty());
    m_dataObject->setData(mimeTypeURL, url);
    m_dataObject->setUrlTitle(title);

    // The URL can also be used as plain text.
    m_dataObject->setData(mimeTypeTextPlain, url.string());

    // The URL can also be used as an HTML fragment.
    m_dataObject->setData(mimeTypeTextHTML, urlToMarkup(url, title));
    m_dataObject->setHtmlBaseUrl(url);
}

void ClipboardWiiU::writeRange(Range* selectedRange, Frame* frame)
{
    ASSERT(selectedRange);
    if (!m_dataObject)
         return;

    m_dataObject->setData(mimeTypeTextHTML, createMarkup(selectedRange, 0, AnnotateForInterchange, false, ResolveNonLocalURLs));
    m_dataObject->setHtmlBaseUrl(frame->document()->url());

    String str = frame->editor()->selectedText();

    static const UChar NonBreakingSpaceCharacter = 0xA0;
    static const UChar SpaceCharacter = ' ';
    str.replace(NonBreakingSpaceCharacter, SpaceCharacter);
    m_dataObject->setData(mimeTypeTextPlain, str);
}

bool ClipboardWiiU::hasData()
{
    if (!m_dataObject)
        return false;

    return m_dataObject->hasData();
}

#if ENABLE(DATA_TRANSFER_ITEMS)
PassRefPtr<DataTransferItems> ClipboardWiiU::items()
{
    RefPtr<DataTransferItemsWiiU> items = DataTransferItemsWiiU::create(this, m_frame->document()->scriptExecutionContext());

    return items;
}
#endif

}
