/*

	Paired Single Implementation of RDFT Code

*/


#include <stdlib.h>
#include <math.h>
#include "libavutil/mathematics.h"
#include "../rdft.h"

#include <ppc.h>
#include <ppc_ps.h>
#include <wkaudio/wkaudAudio.h>

// --------------------------------------------------------------------------------

void ff_fft_permute_ps(FFTContext *s, FFTComplex *z, FFTComplex *tmp, int bInPlace);
void ff_fft_permute_fromseparate_ps(FFTContext *s, const FFTSample *r, const FFTSample *i, FFTComplex *tmp);
void ff_fft_permute_2ndzero_ps(FFTContext *s, FFTComplex *z, FFTComplex *tmp);
void ff_fft_calc_ps(FFTContext *s, FFTComplex *z);

// --------------------------------------------------------------------------------

static const f32x2 cnst_m1p1 = {-1.0f,  1.0f};
static const f32x2 cnst_p1m1 = { 1.0f, -1.0f};

// --------------------------------------------------------------------------------

/** Map one real FFT into two parallel real even and odd FFTs. Then interleave
 * the two real FFTs into one complex FFT. Unmangle the results.
 * ref: http://www.engineeringproductivitytools.com/stuff/T0001/PT10.HTM
 */
// Process from data to dest buffer
// (they may be identical & dest may be NULL - in which case the data stays in a temp buffer returned by the function)
/*
	Important note about buffer usage!!!

	A forward transform with leaving data in the temp buffer will return a pointer to "tmp1".
	The first operation of an inverse will process from the source to "tmp".

	Hence we can keep the results of a FFT inside LC and then feed it back from tmp with no harm!

	Fragile may be... but fast!
*/
FFTSample *ff_rdft_calc_to_ps(RDFTContext* s, const FFTSample* data, FFTSample* dest, int secondZero)
{
    int i, i1, i2, c;
    FFTComplex ev, od;
    const int n = 1 << s->nbits;
    const float k1 = 0.5f;
    const float k2 = 0.5f - s->inverse;
    const FFTSample *tcos = s->tcos;
    const FFTSample *tsin = s->tsin;
	const int bufferSize = ((1 << s->fft.nbits) * sizeof(FFTComplex) + 31) & ~31;

	void *pLC = wksupAudioLCBeginUse(2 * bufferSize);
	FFTComplex *tmp, *tmp1;
	tmp = pLC;
	tmp1 = (FFTComplex *)(((uint32_t)pLC) + bufferSize);

    if (!s->inverse)
	{
		// permute data & keep it in tmp buffer
		if (secondZero)
			ff_fft_permute_2ndzero_ps(&s->fft, (FFTComplex*)data, tmp);
		else
			ff_fft_permute_ps(&s->fft, (FFTComplex*)data, tmp, 0);
		// perform FFT calcs (in place)
		ff_fft_calc_ps(&s->fft, tmp);
    }

	f32x2 vk1 = { k1, -k1};
	f32x2 vk2 = {-k2,  k2};

	// RDFT calcs

	const FFTSample *rd;
	FFTSample *wrt;
	// Are we in inverse mode?
	if (s->inverse)
	{
		// Yes. Read from memory into LC...
		rd = data;
		wrt = (FFTSample*)tmp;		// note: assumes float sample overlaps the real part of a complex number
	}
	else
	{
		// No. Read from LC & write to memory...
		rd = tmp;
		wrt = dest ? dest : (FFTSample*)tmp;	// note: assumes float sample overlaps the real part of a complex number
	}
	
    /* i=0 is a special case because of packing, the DC term is real, so we
       are going to throw the N/2 term (also real) in with it. */
    ev.re = rd[0];
    wrt[0] = ev.re + rd[1];
    wrt[1] = ev.re - rd[1];

	i = 1;
	i1 = 2;
	i2 = n - 2;
	for(c=(n>>2)-1; c>0; --c)
	{
		f32x2 di1 = __PSQ_L(&rd[i1], 0, 0);
		f32x2 di2 = __PSQ_L(&rd[i2], 0, 0);

		float ts = tsin[i];
		float tc = tcos[i];
		f32x2 ts_mp = __PS_MULS0F(cnst_m1p1, ts);		// -s, s
		f32x2 tc_mp = __PS_MULS0F(cnst_m1p1, tc);		// -c, c

//      ev.re =  k1*(data[i1  ]+data[i2  ]);
//      ev.im =  k1*(data[i1+1]-data[i2+1]);
		f32x2 vev = __PS_MADDS0F(di1, k1, __PS_MUL(di2, vk1));


//		od.re =  k2*(data[i1+1]+data[i2+1]);
//		od.im = -k2*(data[i1  ]-data[i2  ]);
		f32x2 vod_rev = __PS_MUL(__PS_ADD(di1, di2), vk2);
		f32x2 vod = __PS_MERGE10(vod_rev, vod_rev);

//      data[i1  ] =  ev.re + od.re*tcos[i] - od.im*tsin[i];
//      data[i1+1] =  ev.im + od.im*tcos[i] + od.re*tsin[i];
		__PSQ_ST(&wrt[i1], __PS_MADD(vod_rev, ts_mp, __PS_MADDS0F(vod, tc, vev)),  0, 0);

//		data[i2  ] =  ev.re - od.re*tcos[i] + od.im*tsin[i];
//		data[i2+1] = -ev.im + od.im*tcos[i] + od.re*tsin[i];
		__PSQ_ST(&wrt[i2], __PS_MADDS0F(vod_rev, ts, __PS_MADD(vod, tc_mp, __PS_MUL(vev, cnst_p1m1))), 0, 0);

		++i;
		i1 += 2;
		i2 -= 2;
    }
	wrt[i1] = rd[i1];
    wrt[i1 + 1] = s->sign_convention * rd[i1 + 1];

	FFTSample *target = wrt;

    if (s->inverse)
	{
        wrt[0] = wrt[0] * k1;
        wrt[1] = wrt[1] * k1;

		// Permutate from primary tmp to secondary tmp1 buffer...
		ff_fft_permute_ps(&s->fft, (FFTComplex*)wrt, tmp1, 0);
		// Do FFT calcs in-place inside secondary tmp1 buffer
		ff_fft_calc_ps(&s->fft, tmp1);

		if (!dest)
			target = (FFTSample*)tmp1;
		else
		{
			// Get data into destination...
			FFTComplex *pD = (FFTComplex*)dest;		// these structures map onto each other, ignoring the imaginary part.
			FFTComplex *pS = (FFTComplex*)tmp1;
			for(int i=(1 << s->fft.nbits); i>0; --i)
				__PSQ_ST(pD++, __PSQ_L(pS++, 0, 0), 0, 0);

			target = dest;
		}
    }

	wksupAudioLCEndUse();
	return(target);
}

// Process in place
static void ff_rdft_calc_ps(RDFTContext* s, FFTSample* data)
{
	ff_rdft_calc_to_ps(s, data, data, 0);
}

// --------------------------------------------------------------------------------


FFTSample *ff_rdft_calc_fromseparate_to_ps(RDFTContext* s, const FFTSample* real, const FFTSample* imag, FFTSample* dest)
{
    int i, i1, i2, c;
    FFTComplex ev, od;
    const int n = 1 << s->nbits;
    const float k1 = 0.5f;
    const float k2 = 0.5f - s->inverse;
    const FFTSample *tcos = s->tcos;
    const FFTSample *tsin = s->tsin;
	const int bufferSize = ((1 << s->fft.nbits) * sizeof(FFTComplex) + 31) & ~31;

	void *pLC = wksupAudioLCBeginUse(2 * bufferSize);
	FFTComplex *tmp, *tmp1;
	tmp = pLC;
	tmp1 = (FFTComplex *)(((uint32_t)pLC) + bufferSize);

	f32x2 vk1 = { k1, -k1};
	f32x2 vk2 = {-k2,  k2};

	FFTSample *wrt;
	FFTSample *target;
	
    if (!s->inverse)
	{
		// permute data & keep it in tmp buffer
		ff_fft_permute_fromseparate_ps(&s->fft, real, imag, tmp);
		// perform FFT calcs (in place)
		ff_fft_calc_ps(&s->fft, tmp);

		//
		// Reading from interleaved internal buffer
		// (!s->inverse)
		//

		const FFTSample *rd = (const FFTSample*)tmp;	// note: assumes float sample overlaps the real part of a complex number
		wrt = dest ? dest : (FFTSample*)tmp;

		/* i=0 is a special case because of packing, the DC term is real, so we
		   are going to throw the N/2 term (also real) in with it. */
		ev.re = rd[0];
		wrt[0] = ev.re + rd[1];
		wrt[1] = ev.re - rd[1];
	
		i = 1;
		i1 = 2;
		i2 = n - 2;
		for(c=(n>>2)-1; c>0; --c)
		{
			f32x2 di1 = __PSQ_L(&rd[i1], 0, 0);
			f32x2 di2 = __PSQ_L(&rd[i2], 0, 0);
	
			float ts = tsin[i];
			float tc = tcos[i];
			f32x2 ts_mp = __PS_MULS0F(cnst_m1p1, ts);		// -s, s
			f32x2 tc_mp = __PS_MULS0F(cnst_m1p1, tc);		// -c, c
	
	//      ev.re =  k1*(data[i1  ]+data[i2  ]);
	//      ev.im =  k1*(data[i1+1]-data[i2+1]);
			f32x2 vev = __PS_MADDS0F(di1, k1, __PS_MUL(di2, vk1));
	
	
	//		od.re =  k2*(data[i1+1]+data[i2+1]);
	//		od.im = -k2*(data[i1  ]-data[i2  ]);
			f32x2 vod_rev = __PS_MUL(__PS_ADD(di1, di2), vk2);
			f32x2 vod = __PS_MERGE10(vod_rev, vod_rev);
	
	//      data[i1  ] =  ev.re + od.re*tcos[i] - od.im*tsin[i];
	//      data[i1+1] =  ev.im + od.im*tcos[i] + od.re*tsin[i];
			__PSQ_ST(&wrt[i1], __PS_MADD(vod_rev, ts_mp, __PS_MADDS0F(vod, tc, vev)),  0, 0);
	
	//		data[i2  ] =  ev.re - od.re*tcos[i] + od.im*tsin[i];
	//		data[i2+1] = -ev.im + od.im*tcos[i] + od.re*tsin[i];
			__PSQ_ST(&wrt[i2], __PS_MADDS0F(vod_rev, ts, __PS_MADD(vod, tc_mp, __PS_MUL(vev, cnst_p1m1))), 0, 0);
	
			++i;
			i1 += 2;
			i2 -= 2;
		}
		wrt[i1] = rd[i1];
		wrt[i1 + 1] = s->sign_convention * rd[i1 + 1];

		target = wrt;
    }
    else
	{
		//
		// Reading from inputs...
		// (s->inverse)
		//
		wrt = tmp;

		/* i=0 is a special case because of packing, the DC term is real, so we
		   are going to throw the N/2 term (also real) in with it. */
		ev.re = real[0];
		wrt[0] = ev.re + imag[0];
		wrt[1] = ev.re - imag[0];
	
		i = 1;
		i1 = 2;
		i2 = n - 2;
		for(c=(n>>2)-1; c>0; --c)
		{
			f32x2 real1 = __PSQ_L(&real[i], 1, 0);
			f32x2 imag1 = __PSQ_L(&imag[i], 1, 0);
			f32x2 real2 = __PSQ_L(&real[i2 >> 1], 1, 0);
			f32x2 imag2 = __PSQ_L(&imag[i2 >> 1], 1, 0);
	
			float ts = tsin[i];
			float tc = tcos[i];
			f32x2 ts_mp = __PS_MULS0F(cnst_m1p1, ts);		// -s, s
			f32x2 tc_mp = __PS_MULS0F(cnst_m1p1, tc);		// -c, c
	
			f32x2 di1 = __PS_MERGE00(real1, imag1);
			f32x2 di2 = __PS_MERGE00(real2, imag2);
	
	//      ev.re =  k1*(data[i1  ]+data[i2  ]);
	//      ev.im =  k1*(data[i1+1]-data[i2+1]);
			f32x2 vev = __PS_MADDS0F(di1, k1, __PS_MUL(di2, vk1));
	
	
	//		od.re =  k2*(data[i1+1]+data[i2+1]);
	//		od.im = -k2*(data[i1  ]-data[i2  ]);
			f32x2 vod_rev = __PS_MUL(__PS_ADD(di1, di2), vk2);
			f32x2 vod = __PS_MERGE10(vod_rev, vod_rev);
	
	//      data[i1  ] =  ev.re + od.re*tcos[i] - od.im*tsin[i];
	//      data[i1+1] =  ev.im + od.im*tcos[i] + od.re*tsin[i];
			__PSQ_ST(&wrt[i1], __PS_MADD(vod_rev, ts_mp, __PS_MADDS0F(vod, tc, vev)),  0, 0);
	
	//		data[i2  ] =  ev.re - od.re*tcos[i] + od.im*tsin[i];
	//		data[i2+1] = -ev.im + od.im*tcos[i] + od.re*tsin[i];
			__PSQ_ST(&wrt[i2], __PS_MADDS0F(vod_rev, ts, __PS_MADD(vod, tc_mp, __PS_MUL(vev, cnst_p1m1))), 0, 0);
	
			++i;
			i1 += 2;
			i2 -= 2;
		}
		wrt[i1] = real[i];
		wrt[i1 + 1] = s->sign_convention * imag[i];

        wrt[0] = wrt[0] * k1;
        wrt[1] = wrt[1] * k1;

		// Permutate from primary tmp to secondary tmp1 buffer...
		ff_fft_permute_ps(&s->fft, (FFTComplex*)wrt, tmp1, 0);
		// Do FFT calcs in-place inside secondary tmp1 buffer
		ff_fft_calc_ps(&s->fft, tmp1);

		if (!dest)
			target = (FFTSample*)tmp1;
		else
		{
			// Get data into destination...
			FFTComplex *pD = (FFTComplex*)dest;		// note: assumes float sample overlaps the real part of a complex number
			FFTComplex *pS = (FFTComplex*)tmp1;
			for(int i=(1 << s->fft.nbits); i>0; --i)
				__PSQ_ST(pD++, __PSQ_L(pS++, 0, 0), 0, 0);

			target = dest;
		}
	}

	wksupAudioLCEndUse();
	return(target);
}

// --------------------------------------------------------------------------------

void ff_rdft_init_ps(RDFTContext *s)
{
    s->rdft_calc = ff_rdft_calc_ps;
}

